r"""
Schemas (:mod:`dasi.schemas`)
=============================

Validated JSON Schemas (https://json-schema.org/)

Check out the :ref:`JSON Schema page <json_schema>` for more information.
"""
import json
from os.path import abspath
from os.path import dirname
from os.path import join
from typing import Dict

import jsonschema

schema_directory = dirname(abspath(__file__))  #: path to directories containing schemas


def _load(filename: str) -> Dict:
    with open(join(schema_directory, filename)) as f:
        return json.load(f)


class Schemas:
    """Class that contains JSON schemas."""

    output_schema = _load("output_schema.json")
    cost_parameters_schema = _load("parameter_json_schema.json")


def validate_with_schema(
    data: Dict, schema: Dict, do_raise: bool = True, reraise_as: Exception = None
):
    """Validate JSON data using a JSON Schema."""
    try:
        jsonschema.validate(data, schema)
    except jsonschema.ValidationError as e:
        if do_raise:
            if reraise_as:
                raise reraise_as(str(e)) from e
            else:
                raise e
        else:
            return False
    return True



r"""
Command Line (:mod:`dasi.command_line`)
=============================

.. currentmodule:: dasi.command_line

This module provide command line interface for DASi.

.. code-block::

    dasi run
"""
import os

import fire
from Bio import BiopythonParserWarning
from Bio import SeqIO
from pyblast.utils import load_fasta_glob
from pyblast.utils import load_genbank_glob
from pyblast.utils import make_circular
from pyblast.utils import make_linear

from dasi import __version__
from dasi import Design
from dasi.cost import SpanCost
from dasi.log import logger


class DasiCLI:
    """DASi command line interface."""

    def __init__(
        self,
        directory=os.getcwd(),
        primers="primers.fasta",
        templates="templates/*.gb",
        fragments="fragments/*.gb",
        goals="goals/*.gb",
        verbose="v",
    ):
        """Initialize a new design."""
        self._directory = directory
        self._primers = os.path.join(self._directory, primers)
        self._templates = os.path.join(self._directory, templates)
        self._fragments = os.path.join(self._directory, fragments)
        self._goals = os.path.join(self._directory, goals)
        self._do_save = True
        self._logger = logger(self)
        if verbose == "v":
            self._logger.set_level("INFO")
        elif verbose == "vv":
            logger.set_level("INFO")
        elif verbose == "vvv":
            logger.set_level("DEBUG")
        elif verbose is None:
            pass
        else:
            raise ValueError(
                "Verbose level '{}' not recognized. " "Select from 'v', 'vv', or 'vvv'"
            )

    def run(self, n_jobs: int = 10):
        """Run a design job.

        :param n_jobs: number of parrallel jobs to run. (default: 10)
        :return:
        """
        import warnings

        warnings.simplefilter(action="ignore", category=RuntimeWarning)
        warnings.simplefilter(action="ignore", category=BiopythonParserWarning)

        self._logger.info("Loading sequence files")
        primers = make_linear(load_fasta_glob(self._primers))
        templates = make_circular(load_genbank_glob(self._templates))
        fragments = make_linear(load_genbank_glob(self._fragments))
        goals = make_circular(load_genbank_glob(self._goals))
        design = Design()
        design.n_jobs = n_jobs
        design.add_materials(
            primers=primers, templates=templates, fragments=fragments, queries=goals
        )

        self._logger.info("Getting span cost model")
        span_cost = self._get_span_cost()
        design.span_cost = span_cost

        self._logger.info("Compiling possible molecular assemblies")
        design.compile()

        self._logger.info("Optimizing molecular assemblies")
        design.optimize()

        self._logger.info("Designing assembly primers and fragments")
        df, adf, design_json = design.to_df()
        adf.to_csv("summary.csv")
        df.to_csv("sequence_design.csv")

        records = []
        for result in design.results.values():
            if result.assemblies:
                a = result.assemblies[0]
                for i, role, m in a.molecules:
                    records.append(m.sequence)

        SeqIO.write(records, os.path.join(self._directory, "sequences.gb"), "genbank")

    def version(self):
        """Print the package version."""
        print(__version__)

    def _get_span_cost(self):
        """Saves the span cost as bytes; reloads when called."""
        path = os.path.join(self._directory, "span_cost.b")
        if self._do_save and os.path.isfile(path):
            with logger.timeit("INFO", "loading bytes"):
                print("Loading file: {}".format(path))
                span_cost = SpanCost.load(path)
        else:
            span_cost = SpanCost.open()
            if self._do_save:
                with logger.timeit("INFO", "saving bytes"):
                    print("Saving file: {}".format(path))
                    span_cost.dump(path)
        return span_cost


def main():
    fire.Fire(DasiCLI)



from .__version__ import __authors__
from .__version__ import __homepage__
from .__version__ import __repo__
from .__version__ import __title__
from .__version__ import __version__
from .cost import SpanCost
from .design import Design
from .design import LibraryDesign
from .log import logger



from copy import deepcopy
from typing import Dict
from typing import List
from typing import Tuple

import networkx as nx

from dasi.constants import Constants
from dasi.design import Design
from dasi.design import DesignResult
from dasi.design.graph_builder import AssemblyGraphPostProcessor
from dasi.models import Alignment
from dasi.models import AlignmentContainer
from dasi.models import AlignmentGroup
from dasi.utils import group_by
from dasi.utils import log_metadata
from dasi.utils import sort_with_keys


def overlapping_groups(group_list_a, group_list_b):
    """Get all groups in group_list_b that right-hand overlap with
    group_list_a."""
    group_sort, group_keys = sort_with_keys(
        group_list_b, key=lambda x: x.query_region.a
    )
    tuples = []
    for group_a in group_list_a:
        overlapping = AlignmentContainer.filter_alignments_by_span(
            group_sort,
            group_a.query_region,
            key=lambda p: p.query_region.a,
            end_inclusive=False,
        )

        if group_a in overlapping:
            overlapping.remove(group_a)
        tuples.append((group_a, overlapping))
    return tuples


def add_clusters(design):
    # list of all alignment groups
    all_groups = []
    for container in design.container_factory.containers().values():
        all_groups += container.get_groups_by_types(Constants.SHARED_FRAGMENT)

    # alignment_groups grouped by query_key
    grouped_by_qk = {}
    for g in all_groups:
        grouped_by_qk.setdefault(g.query_key, list())
        grouped_by_qk[g.query_key].append(g)

    # overlapping_by_qk
    overlapping = []
    for qk, groups in list(grouped_by_qk.items())[:]:
        overlapping += overlapping_groups(groups, groups)

    new_alignments = []
    for group_a, group_list in overlapping:
        new = container.expand_overlaps(
            group_list + [group_a], include_left=False, atype=Constants.SHARED_FRAGMENT
        )
        new_alignments += new
    new_alignments = list({a.eq_hash(): a for a in new_alignments}.values())
    design.container_factory.add_alignments(new_alignments)


def to_undirected(graph):
    """.to_undirected is implemented in networkx out of the box, however, it
    suffers from occational infinite recursion errors during the deepcopy phase
    of the method (unknown as to why)."""
    undirected = nx.Graph()
    copied = deepcopy(graph)
    for n in copied.nodes:
        ndata = copied.nodes[n]
        undirected.add_node(n, **ndata)
    for n1, n2 in copied.edges:
        edata = copied.edges[n1, n2]
        undirected.add_edge(n1, n2, **edata)
    return undirected


def get_subgraphs(graph):
    """Get independent subgraphs."""
    node_list = list(graph.nodes)
    subgraphs = []
    while len(node_list) > 0:
        node = node_list[-1]
        subgraph = nx.bfs_tree(to_undirected(graph), node)
        for n in subgraph.nodes:
            node_list.remove(n)
        subgraphs.append(graph.subgraph(subgraph.nodes))
    return subgraphs


def has_repeats(g):
    """Check if the interaction graph has a repeated DNA sequence."""
    grouped_by_key = {}
    for n in g.nodes:
        grouped_by_key.setdefault(n[0], list())
        grouped_by_key[n[0]].append((n[1], n[2]))
    for k, v in grouped_by_key.items():
        if len(v) > 1:
            print(grouped_by_key)
            return True
    return False


def cluster_graph(design):
    interaction_graph = nx.Graph()
    all_groups = []
    for container in design.container_factory.containers().values():
        all_groups += container.get_groups_by_types(Constants.SHARED_FRAGMENT)
    for g in all_groups:
        for a in g.alignments:
            n1 = (a.query_key, a.query_region.a, a.query_region.b, a.query_region.c)
            n2 = (
                a.subject_key,
                a.subject_region.a,
                a.subject_region.b,
                a.subject_region.c,
            )
            edata = interaction_graph.get_edge_data(n1, n2)
            if edata:
                edata.setdefault("alignments", list())
                edata["alignments"].append(a)
            else:
                interaction_graph.add_edge(n1, n2, alignments=[a])
    graphs = get_subgraphs(interaction_graph)
    graphs = [g for g in graphs if not has_repeats(g)]
    graphs.sort(reverse=True, key=lambda x: x.number_of_nodes())
    return graphs


class LibraryDesign(Design):
    """Design class for producing assemblies for libraries."""

    DEFAULT_N_JOBS = 10
    FAVOR_SHARED_SEQUENCES = 2
    DEFAULT_N_ASSEMBLIES = Design.DEFAULT_N_ASSEMBLIES
    ALGORITHM = Constants.ALGORITHM_LIBRARY

    def __init__(self, span_cost=None, seqdb=None, n_jobs=None):
        """Something."""
        super().__init__(span_cost=span_cost, seqdb=seqdb, n_jobs=n_jobs)
        self.shared_alignments = []
        self._edges = []

    # @staticmethod
    # def _get_repeats_from_results(results):
    #     repeats = []
    #     for r in results:
    #         qk = r['query']['origin_key']
    #         sk = r['subject']['origin_key']
    #         if qk == sk:
    #             repeats.append((qk, r['query']['start'], r['query']['end']))
    #     return repeats

    # TODO: why?
    @staticmethod
    def _get_iter_non_repeats(alignments: List[Alignment]) -> Tuple[str, int, int]:
        """Return repeat regions of alignments. These are alignments that align
        to themselves.

        :param alignments:
        :return:
        """
        for align in alignments:
            qk = align.query_key
            sk = align.subject_key
            if qk == sk:
                yield (qk, align.query_region.a, align.query_region.b)

    def _expand_from_synthesized(self):
        """Expand PCR products that are next to SYNTHESIZED_FRAGMENTS.

        :return:
        """
        for query_key, container in self.container_factory.containers().items():

            def is_not_shared(group_a: AlignmentGroup, group_b: AlignmentGroup):
                if group_a.type == Constants.SHARED_SYNTHESIZED_FRAGMENT:
                    return False
                if group_b.type == Constants.SHARED_SYNTHESIZED_FRAGMENT:
                    return True
                return False

            new_alignments = container.expand_overlaps(
                container.get_groups_by_types(
                    [
                        Constants.FRAGMENT,
                        Constants.PCR_PRODUCT,
                        Constants.SHARED_SYNTHESIZED_FRAGMENT,
                    ]
                ),
                Constants.PCR_PRODUCT,
                pass_condition=is_not_shared,
            )
            container.add_alignments(new_alignments, lim_size=True)

            self.logger.info(
                "{}: Expanded {} using {} and found {} new alignments.".format(
                    query_key,
                    Constants.PCR_PRODUCT,
                    Constants.SHARED_SYNTHESIZED_FRAGMENT,
                    len(new_alignments),
                )
            )

            # # grab the pcr products and expand primer pairs (again)
            # templates = container.get_groups_by_types(Constants.PCR_PRODUCT)
            # new_primer_pairs = container.expand_primer_pairs(templates)
            # self.logger.info(
            #     "{}: Expanded {} {} using {}".format(
            #         query_key,
            #         len(new_primer_pairs),
            #         Constants.PCR_PRODUCT,
            #         Constants.SYNTHESIZED_FRAGMENT,
            #     )
            # )

    def _expand_synthesized_fragments(self):
        """
            1. copy groups from SHARED_FRAGMENTS to SYNTHESIZED_FRAGMENT
            2. expand overlaps for SYNTHESIZED_FRAGMENTS
        """
        for query_key, container in self.container_factory.containers().items():
            # expand the share fragments using their own endpoints
            original_shared_fragments = container.get_groups_by_types(
                Constants.SHARED_FRAGMENT
            )
            copied_alignments = container.copy_groups(
                original_shared_fragments, Constants.SHARED_SYNTHESIZED_FRAGMENT
            )
            container.add_alignments(copied_alignments, lim_size=True)

            new_alignments = container.expand_overlaps(
                container.get_groups_by_types(Constants.SHARED_SYNTHESIZED_FRAGMENT),
                atype=Constants.SHARED_SYNTHESIZED_FRAGMENT,
            )

            self.logger.info(
                "{}: Expanded {} new {} alignments.".format(
                    query_key,
                    len(new_alignments),
                    Constants.SHARED_SYNTHESIZED_FRAGMENT,
                )
            )
            container.add_alignments(new_alignments, lim_size=True)

    def _check_shared_repeats(self):
        repeats = []
        for query_key, container in self.container_factory.containers().items():
            for align in container.get_alignments_by_types(Constants.SHARED_FRAGMENT):
                qk = align.query_key
                sk = align.subject_key
                if qk == sk:
                    repeats.append(align)
        assert not repeats

    def _share_query_blast(self):
        """Find and use shared fragments across queries.

        :return:
        """

        # step 1: get query-on-query alignments
        self.logger.info("=== Expanding shared library fragments ===")
        blast = self.blast_factory(self.QUERIES, self.QUERIES)
        blast.blastn()

        results = blast.get_perfect()

        # step 2: eliminate self binding results
        results = [
            entry
            for entry in results
            if entry["query"]["origin_key"] != entry["subject"]["origin_key"]
        ]

        self.logger.info(
            "Found {} shared alignments between the queries".format(len(results))
        )

        # step 3: load results to the container
        self.shared_alignments = results
        self.container_factory.seqdb.update(blast.seq_db.records)
        self.container_factory.load_blast_json(results, Constants.SHARED_FRAGMENT)

    def _precompile_library_expansion(self):
        self._share_query_blast()
        self._expand_synthesized_fragments()
        self._expand_from_synthesized()
        self._check_shared_repeats()

    @log_metadata("compile", additional_metadata={"algorithm": ALGORITHM})
    def compile(self, n_jobs: int = DEFAULT_N_JOBS):
        """Compile the materials list into assembly graphs."""
        self._uncompile()
        tracker = self.logger.track("INFO", desc="Compiling library", total=4).enter()

        self.graphs = {}

        tracker.update(0, "Running blast")
        self._blast()

        tracker.update(1, "Running shared fragment blast")
        self._share_query_blast()

        tracker.update(2, "Finding shared clusters")
        self.update_library_metadata()

        tracker.update(3, "Assembling graphs")
        self.assemble_graphs(n_jobs=n_jobs)

        # TODO: adjust n_clusters in the graph instead

        tracker.update(4, "Post processing")
        adjusted = 0
        for qk, graph in self.graphs.items():
            query = self.seqdb[qk]
            processor = AssemblyGraphPostProcessor(graph, query, self.span_cost)
            processor()
            for n1, n2, edata in graph.edges(data=True):
                if edata["type_def"].name == Constants.SHARED_SYNTHESIZED_FRAGMENT:
                    group = edata["group"]
                    edata["notes"] += "n_clusters: {}".format(group.meta["n_clusters"])
                    # TODO: adjust n_clusters
                    edata["material"] = (
                        edata["material"]
                        / (group.meta["n_clusters"])
                        / self.FAVOR_SHARED_SEQUENCES
                    )
                    edata["cost"] = edata["material"] / edata["efficiency"]
                    adjusted += 1
        tracker.update(4, "Post processing complete")
        tracker.update(4, "Compilation complete")

    def update_library_metadata(self):
        add_clusters(self)
        graphs = cluster_graph(self)

        for c in self.container_list:
            c.share_group_tag = {}

        # update the meta data
        for g in graphs:
            n_clusters = g.number_of_nodes()
            for n1, n2, edata in g.edges(data=True):
                alignments = edata["alignments"]
                for qk, container_alignments in group_by(
                    alignments, key=lambda x: x.query_key
                ).items():
                    container = self.containers[qk]
                    for a in container_alignments:
                        group_key = (
                            a.query_region.a,
                            a.query_region.b,
                            Constants.SHARED_SYNTHESIZED_FRAGMENT,
                        )
                        container.group_tags.add(
                            Constants.SHARE_GROUP_TAG,
                            group_key,
                            {
                                "alignments": container_alignments,
                                "cross_container_alignments": alignments,
                                "n_clusters": n_clusters,
                            },
                        )

        # for container in design.container_list():
        #     groups = container.get_groups_by_types(Constants.SHARED_FRAGMENT)
        #     copied = container.copy_groups(groups, atype=Constants.SYNTHESIZED_FRAGMENT)
        #     container.add_alignments(copied)

        # def is_not_shared(group_a: AlignmentGroup, group_b: AlignmentGroup):
        #     if group_a.type == Constants.SHARED_FRAGMENT:
        #         return False
        #     if group_b.type == Constants.SYNTHESIZED_FRAGMENT:
        #         return True
        #     return False

        # new_alignments = container.expand_overlaps(
        #     container.get_groups_by_types(
        #         [
        #             Constants.FRAGMENT,
        #             Constants.PCR_PRODUCT,
        #             Constants.SYNTHESIZED_FRAGMENT,
        #         ]
        #     ),
        #     Constants.PCR_PRODUCT,
        #     pass_condition=is_not_shared,
        # )
        # container.add_alignments(new_alignments, lim_size=True)

    # def post_process_library_graphs(self):
    #     for qk, graph in self.graphs.items():
    #         query = self.seqdb[qk]
    #         processor = AssemblyGraphPostProcessor(graph, query)
    #         processor()

    @log_metadata(
        "optimize", additional_metadata={"algorithm": Constants.ALGORITHM_LIBRARY}
    )
    def optimize(
        self, n_paths: int = DEFAULT_N_ASSEMBLIES, n_jobs: int = DEFAULT_N_JOBS
    ) -> Dict[str, DesignResult]:
        """Optimize the assembly graph for library assembly."""
        return super().optimize(n_paths=n_paths, n_jobs=n_jobs)



r"""
Models (:mod:`dasi.models`)
=============================

.. currentmodule:: dasi.models

This module provides representations of Molecules and Molecular Assemblies.

.. autosummary::
    :toctree: generated/

    Assembly
    AssemblyNode
    Molecule
    MoleculeType
    Reaction
    Alignment
    AlignmentGroup
    AlignmentGroupBase
    MultiPCRProductAlignmentGroup
    PCRProductAlignmentGroup
    AlignmentContainer
    AlignmentContainerFactory

modules
-------

.. autosummary::
    :toctree: generated/

    assembly
    molecule
    alignment
    alignment_container
"""
from .alignment import Alignment
from .alignment import AlignmentGroup
from .alignment import AlignmentGroupBase
from .alignment import MultiPCRProductAlignmentGroup
from .alignment import PCRProductAlignmentGroup
from .alignment_container import AlignmentContainer
from .alignment_container import AlignmentContainerFactory
from .assembly import Assembly
from .assembly import AssemblyNode
from .molecule import Molecule
from .molecule import MoleculeType
from .molecule import Reaction



import bisect
import itertools
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union
from uuid import uuid4

import networkx as nx
import numpy as np
from Bio.SeqRecord import SeqRecord
from more_itertools import partition
from pyblast.utils import is_circular

from dasi.constants import Constants
from dasi.cost import SpanCost
from dasi.log import logger
from dasi.models import AlignmentContainer
from dasi.models import AlignmentGroup
from dasi.models import AssemblyNode
from dasi.models import MoleculeType
from dasi.models import PCRProductAlignmentGroup
from dasi.utils import bisect_between
from dasi.utils import Region
from dasi.utils import sort_with_keys
from dasi.utils.sequence_complexity import DNAStats


class AssemblyGraphBuilder:
    """Class that builds an AssemblyGraph from an alignment container."""

    COST_THRESHOLD = 10000

    def __init__(self, alignment_container: AlignmentContainer, span_cost=None):
        self.container = alignment_container
        if span_cost is None:
            self.span_cost = SpanCost.open()
        else:
            self.span_cost = span_cost
        self.G = None
        self.logger = logger(self)

    def add_node(self, node: AssemblyNode) -> None:
        """Add node to the graph.

        :param node: the assembly node to add
        :return: None
        """
        self.G.add_node(AssemblyNode(*node))

    @staticmethod
    def add_edge_to_graph(
        graph,
        n1: AssemblyNode,
        n2: AssemblyNode,
        cost: Union[float, None],
        material: Union[float, None],
        time: Union[float, None],
        efficiency: Union[float, None],
        span: int,
        atype: MoleculeType,
        group: Union[AlignmentGroup, PCRProductAlignmentGroup],
        notes: str = "",
    ):
        """Add an edge between two assembly nodes.

        :param n1: src node
        :param n2: dest node
        :param name: name of the edge
        :param cost: overall cost of the edge
        :param material: material cost of the edge. Used in path calculations.
        :param time: time cost of the edge. Used in path calculations.
        :param efficiency: efficiency of the edge. Used in path calculations.
        :param span: spanning distance (in bp) of the edge.
        :param atype: alignment type of the edge.
        :param kwargs: additional kwargs for the edge data
        :return:
        """
        return graph.add_edge(
            n1,
            n2,
            cost=cost,
            material=material,
            time=time,
            efficiency=efficiency,
            span=span,
            type_def=atype,
            group=group,
            notes=notes,
        )

    def add_edge(
        self,
        n1: AssemblyNode,
        n2: AssemblyNode,
        cost: Union[float, None],
        material: Union[float, None],
        time: Union[float, None],
        efficiency: Union[float, None],
        span: int,
        atype: MoleculeType,
        internal_or_external: str,
        condition: Tuple[bool, bool],
        group: Union[AlignmentGroup, PCRProductAlignmentGroup],
    ):
        """Add an edge between two assembly nodes.

        :param n1: src node
        :param n2: dest node
        :param name: name of the edge
        :param cost: overall cost of the edge
        :param material: material cost of the edge. Used in path calculations.
        :param time: time cost of the edge. Used in path calculations.
        :param efficiency: efficiency of the edge. Used in path calculations.
        :param span: spanning distance (in bp) of the edge.
        :param atype: alignment type of the edge.
        :param kwargs: additional kwargs for the edge data
        :return:
        """

        assert condition == atype.design
        assert internal_or_external == atype.int_or_ext
        return self.add_edge_to_graph(
            graph=self.G,
            n1=n1,
            n2=n2,
            cost=cost,
            material=material,
            time=time,
            efficiency=efficiency,
            span=span,
            atype=atype,
            group=group,
            notes="",
        )

    def iter_internal_edge_data(
        self, group: Union[AlignmentGroup, PCRProductAlignmentGroup]
    ) -> dict:
        q = group.query_region

        mtype = MoleculeType.types[group.type]
        a_expand, b_expand = mtype.design
        internal_cost = mtype.cost
        internal_efficiency = mtype.efficiency

        if q.cyclic:
            # cyclic
            if q.b < q.a:
                pairs = [(q.a, q.b + q.context_length), (q.a + q.context_length, q.b)]
            else:
                pairs = [(q.a, q.b), (q.a + q.context_length, q.b + q.context_length)]
        else:
            # linear
            pairs = [(q.a, q.b)]

        # nodes = []
        # edges = []
        for a, b in pairs:
            if a is not None:
                anode = (a, a_expand, "A")
                # nodes.append(anode)
                if b is not None:
                    bnode = (b, b_expand, "B")
                    # nodes.append(bnode)
                    yield (
                        anode,
                        bnode,
                        dict(
                            material=internal_cost,
                            cost=internal_cost / internal_efficiency,
                            time=0.1,
                            internal_or_external="internal",
                            span=len(group.query_region),
                            atype=MoleculeType.types[group.type],
                            efficiency=internal_efficiency,
                            condition=(a_expand, b_expand),
                            group=group,
                        ),
                    )
                    # edges.append(edge)
        # return nodes, edges

    def add_internal_edges(
        self, groups: List[Union[AlignmentGroup, PCRProductAlignmentGroup]]
    ):
        for g in groups:
            for a, b, ab_data in self.iter_internal_edge_data(g):
                for a_overhang, b_overhang in itertools.product(
                    [True, False], repeat=2
                ):
                    a_node = AssemblyNode(a[0], a[1], a[2], a_overhang)
                    b_node = AssemblyNode(b[0], b[1], b[2], b_overhang)
                    self.add_edge(a_node, b_node, **ab_data)

    def add_external_edges(self, groups, group_keys, nodes: Iterable[AssemblyNode]):
        if not groups:
            return
        query_region = groups[0].query_region
        length = query_region.context_length
        a_nodes, b_nodes = partition(lambda x: x.type == "B", nodes)

        a_nodes = sorted(a_nodes, key=lambda x: x.index)
        b_nodes = sorted(b_nodes, key=lambda x: x.index)

        a_nodes_gap, a_nodes_overhang = [
            list(x) for x in partition(lambda x: x.overhang, a_nodes)
        ]
        b_nodes_gap, b_nodes_overhang = [
            list(x) for x in partition(lambda x: x.overhang, b_nodes)
        ]

        def make_overlap_iterator(a, b):
            """Find all nodes that satisfy the below condition: b.

            |--------|
                   |-------|
                   a

            With the exception that when a.index == b.index, this is not
            considered an overlap, but covered in the `make_gap_iterator`,
            due to using bisect.bisect_right. Overlap is more computationally
            intensive.
            """
            a, akeys = sort_with_keys(a, lambda x: x.index)
            b, bkeys = sort_with_keys(b, lambda x: x.index)
            for _a in a:
                i = bisect.bisect_right(bkeys, _a.index)
                for _b in b[i:]:
                    yield _b, _a

        def make_gap_itererator(a, b):
            """Find all nodes that satisfy the below condition: b.

            |--------|              |-------|              a
            """
            a, akeys = sort_with_keys(a, lambda x: x.index)
            b, bkeys = sort_with_keys(b, lambda x: x.index)

            for _a in a:
                i = bisect.bisect_right(bkeys, _a.index)
                for _b in b[:i]:
                    yield _b, _a

        def make_origin_iterator(a, b):
            a, akeys = sort_with_keys(a, lambda x: x.index)
            b, bkeys = sort_with_keys(b, lambda x: x.index)

            i = bisect.bisect_right(akeys, length)
            j = bisect.bisect_left(bkeys, length)
            for _a in a[:i]:
                for _b in b[j:]:
                    yield _b, _a

        overlap_iter = make_overlap_iterator(a_nodes_overhang, b_nodes_overhang)
        gap_iter = make_gap_itererator(a_nodes_gap, b_nodes_gap)
        gap_origin_iter = make_origin_iterator(a_nodes_gap, b_nodes_gap)
        overlap_origin_iter = make_origin_iterator(a_nodes_overhang, b_nodes_overhang)

        for bnode, anode in overlap_iter:
            self.add_overlap_edge(bnode, anode, query_region, group_keys, groups)

        for bnode, anode in gap_iter:
            self.add_gap_edge(bnode, anode, query_region)

        for bnode, anode in overlap_origin_iter:
            self.add_overlap_edge(
                bnode, anode, query_region, group_keys, groups, origin=True
            )
        for bnode, anode in gap_origin_iter:
            self.add_gap_edge(bnode, anode, query_region, origin=True)

    def add_overlap_edge(
        self, bnode, anode, query_region, group_keys, groups, origin=False
    ):
        q = query_region.new(anode.index, bnode.index)
        if len(q) == 0:
            return
        if not origin:
            i, j = bisect_between(group_keys, q.a, q.b)
            filtered_groups = groups[i:j]
        else:
            i = bisect.bisect_left(group_keys, q.a)
            j = bisect.bisect_right(group_keys, q.b)
            filtered_groups = groups[i:] + groups[:j]
        if filtered_groups:
            if not origin:
                span = anode.index - bnode.index
            else:
                span = -len(q)
            if span <= 0:
                condition = (bnode.expandable, anode.expandable)
                self.add_edge(
                    bnode,
                    anode,
                    cost=None,
                    material=None,
                    time=None,
                    efficiency=None,
                    internal_or_external="external",
                    atype=MoleculeType.types[Constants.OVERLAP](condition),
                    condition=condition,
                    span=span,
                    group=None,
                )

    def add_gap_edge(self, bnode, anode, query_region, origin=False):
        try:
            if not origin:
                span = anode.index - bnode.index
            else:
                if bnode.index == query_region.context_length:
                    span = len(query_region.new(0, anode.index))
                else:
                    span = len(query_region.new(bnode.index, anode.index))
        except Exception as e:
            print(
                "An exception was raised while attempting to create a abstract region"
            )
            print("origin == " + str(origin))
            print("bnode == " + str(bnode))
            print("anode == " + str(anode))
            print("query_region == " + str(query_region))
            raise e
        if span >= 0:
            condition = (bnode.expandable, anode.expandable)
            # cost_dict = self._get_cost(span, condition)
            self.add_edge(
                bnode,
                anode,
                cost=None,
                material=None,
                time=None,
                efficiency=None,
                internal_or_external="external",
                atype=MoleculeType.types[Constants.GAP](condition),
                span=span,
                condition=condition,
                group=None,
            )

    @staticmethod
    def batch_add_edge_costs(graph, edges, span_cost, cost_threshold: float = None):
        """Add costs to all edges at once.

        :return:
        """
        # add external_edge_costs
        edge_dict = {}
        for n1, n2, edata in edges:
            condition = edata["type_def"].design
            assert isinstance(condition, tuple)
            edge_dict.setdefault(condition, []).append(((n1, n2), edata, edata["span"]))

        edges_to_remove = []
        for condition, info in edge_dict.items():
            edges, edata, spans = zip(*info)
            npdf = span_cost.cost(np.array(spans), condition)
            data = npdf.aggregate(np.vstack)
            cost_i = [i for i, col in enumerate(npdf.columns) if col == "cost"][0]

            for i, (e, edge) in enumerate(zip(edata, edges)):
                if cost_threshold is not None and data[cost_i, i] > cost_threshold:
                    edges_to_remove.append(edge)
                else:
                    _d = {c: data[col_i, i] for col_i, c in enumerate(npdf.columns)}
                    e.update(_d)
        if cost_threshold is not None:
            graph.remove_edges_from(edges_to_remove)

    def _get_cost(self, bp, ext):
        data = self.span_cost(bp, ext).data
        return {k: v[0] for k, v in data.items()}

    def build_assembly_graph(self) -> nx.DiGraph:

        self.G = nx.DiGraph(name="Assembly Graph")

        groups, group_keys = sort_with_keys(
            self.container.get_groups_by_types(
                [
                    Constants.PCR_PRODUCT,
                    Constants.PCR_PRODUCT_WITH_RIGHT_PRIMER,
                    Constants.PCR_PRODUCT_WITH_LEFT_PRIMER,
                    Constants.PCR_PRODUCT_WITH_PRIMERS,
                    Constants.PRIMER_EXTENSION_PRODUCT_WITH_PRIMERS,
                    Constants.PRIMER_EXTENSION_PRODUCT_WITH_RIGHT_PRIMER,
                    Constants.PRIMER_EXTENSION_PRODUCT_WITH_LEFT_PRIMER,
                    Constants.FRAGMENT,
                    Constants.SHARED_SYNTHESIZED_FRAGMENT,
                ]
            ),
            key=lambda g: g.query_region.a,
        )

        self.add_internal_edges(groups)
        self.add_external_edges(groups, group_keys, self.G.nodes())

        edges = []
        for n1, n2, edata in self.G.edges(data=True):
            if edata["cost"] is None:
                edges.append((n1, n2, edata))
            elif edata["type_def"].name == Constants.SHARED_SYNTHESIZED_FRAGMENT:
                edges.append((n1, n2, edata))
        self.batch_add_edge_costs(self.G, edges, self.span_cost, self.COST_THRESHOLD)

        # TODO: freeze?
        # nx.freeze(self.G)
        return self.G


# TODO: evaluate primer designs, scoring PCR products d(how long does this take?)
class AssemblyGraphPostProcessor:
    """Post-processing for assembly graphs. Evaluates:

    1. synthesis complexity and weights corresponding edge
    2. pcr product efficiency
    3. (optional) optimal partitions for synthesis fragments
    """

    def __init__(self, graph: nx.DiGraph, query: SeqRecord, span_cost: SpanCost):
        self.graph = graph
        self.query = query
        self.stats = DNAStats(
            self.query + self.query,
            repeat_window=14,
            stats_window=20,
            hairpin_window=20,
        )
        self.logged_msgs = []
        # TODO: make a more sophisticated complexity function?
        # TODO: expose this to input parameters
        self.COMPLEXITY_THRESHOLD = 10.0
        self.logger = logger(self)
        self.span_cost = span_cost

    @staticmethod
    def optimize_partition(
        signatures: np.ndarray, step: int, i: int = None, j: int = None
    ):
        """Optimize partition by minimizing the number of signatures in the
        given array.

        :param signatures: array of signatures
        :param step: step size
        :param i:
        :param j:
        :return:
        """
        d = []

        if i is None:
            i = 0
        if j is None:
            j = signatures.shape[1]

        for x in range(i, j, step):
            m1 = np.empty(signatures.shape[1])
            m2 = m1.copy()
            m1.fill(np.nan)
            m2.fill(np.nan)

            m1[:x] = np.random.uniform(1, 10)
            m2[x:] = np.random.uniform(1, 10)

            d += [m1, m2]
        d = np.vstack(d)
        z = np.tile(d, signatures.shape[0]) * signatures.flatten()

        partition_index = np.repeat(
            np.arange(0, signatures.shape[1], step),
            signatures.shape[0] * signatures.shape[1] * 2,
        )

        a, b, c = np.unique(z, return_counts=True, return_index=True)
        i = b[np.where(c > 1)]
        a, c = np.unique(partition_index[i], return_counts=True)
        if len(c):
            arg = c.argmin()
            return a[arg], c[arg]

    def _edge_to_region(self, n1, n2):
        if n1.index == len(self.query) and is_circular(self.query):
            a = 0
            b = n2.index
        else:
            a = n1.index
            b = n2.index
        return Region(a, b, len(self.query), cyclic=is_circular(self.query))

    def _complexity_to_efficiency(self, edata):
        if edata["complexity"] > self.COMPLEXITY_THRESHOLD:
            edata["efficiency"] = 0.1
            return True
        return False

    def update_complexity(self, n1, n2, edata):
        bad_edges = []
        if edata["type_def"].synthesize:
            span = edata["span"]
            if span > 0:
                # TODO: cyclic may not always be tru
                r = self._edge_to_region(n1, n2)
                complexity = self.stats.cost(r.a, r.c)
                edata["complexity"] = complexity
                if self._complexity_to_efficiency(edata):
                    bad_edges.append((n1, n2, edata))
        return bad_edges

    def update_long_pcr_products(self, n1, n2, edata):
        if edata["type_def"].name in [
            Constants.PCR_PRODUCT,
            Constants.PCR_PRODUCT_WITH_LEFT_PRIMER,
            Constants.PCR_PRODUCT_WITH_RIGHT_PRIMER,
            Constants.PCR_PRODUCT_WITH_PRIMERS,
        ]:
            span = edata["span"]
            if span > 4000:
                edata["efficiency"] *= 0.8
            elif span > 5000:
                edata["efficiency"] *= 0.5

    def synthesis_partitioner(self, n1, n2, edata, border):
        r = self._edge_to_region(n1, n2)
        partitions = self.stats.partition(
            10,
            25,
            i=r.a,
            j=r.c,
            border=border,
            stopping_threshold=self.COMPLEXITY_THRESHOLD,
        )

        if not partitions:
            return []

        best_partition = partitions[0]
        n3 = AssemblyNode(
            best_partition["index_1"][1], False, str(uuid4()), overhang=True
        )
        n4 = AssemblyNode(
            best_partition["index_2"][0], False, str(uuid4()), overhang=True
        )

        AssemblyGraphBuilder.add_edge_to_graph(
            self.graph,
            n1,
            n3,
            cost=None,
            material=None,
            time=None,
            efficiency=None,
            span=len(self._edge_to_region(n1, n3)),
            atype=edata["type_def"],
            group=edata["group"],
        )

        # internal edge
        AssemblyGraphBuilder.add_edge_to_graph(
            self.graph,
            n3,
            n4,
            cost=None,
            material=None,
            time=None,
            efficiency=None,
            span=-len(self._edge_to_region(n4, n3)),
            atype=MoleculeType.types[Constants.OVERLAP],
            group=None,
        )

        # overlap edge
        AssemblyGraphBuilder.add_edge_to_graph(
            self.graph,
            n4,
            n2,
            cost=None,
            material=None,
            time=None,
            efficiency=None,
            span=-len(self._edge_to_region(n4, n2)),
            atype=edata["type_def"],
            group=edata["group"],
        )

        edata1 = self.graph.get_edge_data(n1, n3)
        edata2 = self.graph.get_edge_data(n3, n4)
        edata3 = self.graph.get_edge_data(n4, n2)

        edata1["complexity"] = best_partition["cost_1"]
        edata3["complexity"] = best_partition["cost_2"]
        edata3["notes"] += " Partitioned"
        edata2["notes"] += " Partitioned"
        edata1["notes"] += " Partitioned"

        edge1 = (n1, n3, edata1)
        edge2 = (n3, n4, edata2)
        edge3 = (n4, n2, edata3)
        return [edge1, edge2, edge3]

    def partition(self, edges):
        new_edges = []
        for n1, n2, edata in self.logger.tqdm(
            edges, "INFO", desc="Paritioning sequences..."
        ):
            min_size = (
                edata["type_def"].min_size
                or MoleculeType.types[Constants.SHARED_SYNTHESIZED_FRAGMENT].min_size
            )
            if edata["span"] > min_size * 2:
                new_edges += self.synthesis_partitioner(n1, n2, edata, border=min_size)

        syn_edges = [e for e in new_edges if e[2]["type_def"].design]
        AssemblyGraphBuilder.batch_add_edge_costs(
            self.graph, syn_edges, self.span_cost, None
        )

        for n1, n2, edata in syn_edges:
            self._complexity_to_efficiency(edata)

    def update(self):
        bad_edges = []
        for n1, n2, edata in self.graph.edges(data=True):
            self.update_long_pcr_products(n1, n2, edata)
            bad_edges += self.update_complexity(n1, n2, edata)

        self.logger.info(
            "Found {} highly complex synthesis segments".format(len(bad_edges))
        )

        # self.partition(bad_edges)

    # TODO: add logging to graph post processor
    # TODO: partition gaps
    def __call__(self):
        self.logger.info("Post processing graph for {}".format(self.query.name))
        self.update()



import functools
import hashlib
import operator
from typing import Dict
from typing import Tuple
from typing import Union

import networkx as nx

from dasi.constants import Constants as C
from dasi.exceptions import DasiOutputValidationError
from dasi.models import Molecule
from dasi.models import MoleculeType
from dasi.models import Reaction
from dasi.schemas import Schemas
from dasi.schemas import validate_with_schema
from dasi.utils.biopython import seqrecord_to_json


def validate_output(data, do_raise: bool = True):
    return validate_with_schema(
        data,
        Schemas.output_schema,
        do_raise=do_raise,
        reraise_as=DasiOutputValidationError,
    )


def shash(seq: str) -> str:
    """Convert sequence string into a hash."""
    return hashlib.sha1(seq.strip().upper().encode()).hexdigest()


def mhash(mol1: Molecule) -> Tuple:
    """Convert molecule to hash."""
    return ("molecule", mol1.type.name, shash(str(mol1.sequence.seq)))


def rhash(rxn: Reaction, assembly_id: int) -> Tuple:
    """Convert reaction to hash."""
    outputs = sorted([mhash(m) for m in rxn.outputs])
    inputs = sorted([mhash(m) for m in rxn.inputs])
    if rxn.name == "Assembly":
        x = ["reaction", rxn.name, assembly_id]
    else:
        x = ["reaction", rxn.name]
    return tuple(x + [tuple(inputs)] + [tuple(outputs)])


def dasi_design_to_dag(design: Union["Design", "LibraryDesign"]) -> nx.DiGraph:
    # TODO: standard way to display SeqRecord
    graph = nx.DiGraph()
    for q_i, (qk, result) in enumerate(design.results.items()):
        for a_i, a in enumerate(result.assemblies):
            dk = (qk, a_i)
            for r in a.reactions:
                r1 = rhash(r, a_i)
                graph.add_node(r1, reaction=r, assembly_index=a_i, assembly=a)
                graph.nodes[r1].setdefault("design_keys", set()).add(dk)
                for in_mol in r.inputs:
                    n1 = mhash(in_mol)
                    graph.add_node(n1, molecule=in_mol)
                    graph.nodes[n1].setdefault("design_keys", set()).add(dk)
                    graph.add_edge(n1, r1)
                for out_mol in r.outputs:
                    n2 = mhash(out_mol)
                    graph.add_node(n2, molecule=out_mol)
                    graph.nodes[n2].setdefault("design_keys", set()).add(dk)
                    graph.add_edge(r1, n2)
    _validate_dag_graph(graph)
    return graph


def _validate_dag_graph(graph: nx.DiGraph):
    """Validates the DASi DAG graph."""
    for n2, ndata2 in graph.nodes(data=True):
        ########################
        # validate each non-plasmid is expected to
        # have at most one reaction
        ########################
        molecule = ndata2.get("molecule", None)
        if molecule and molecule.type is not MoleculeType.types[C.PLASMID]:
            predecessors = list(graph.predecessors(n2))
            if len(predecessors) > 1:
                msg = str(n2)
                for n1 in graph.predecessors(n2):
                    msg += "\n\t" + str(n1)
                raise ValueError(
                    "Molecule {} has more than one reaction.\n{}".format(n2, msg)
                )

        ########################
        # validate each reaction
        # has expected number of
        # predecessors and successors
        ########################
        reaction = ndata2.get("reaction", None)
        if reaction:
            predecessors = list(graph.predecessors(n2))
            if not len(predecessors) == len(reaction.inputs):
                raise ValueError(
                    "The number of predecessors ({}) does not"
                    " match the number of inputs for the reaction"
                    " ({})".format(len(predecessors), len(reaction.outputs))
                )

            successors = list(graph.successors(n2))
            if not len(successors) == len(reaction.outputs):
                raise ValueError(
                    "The number of successors ({}) does not"
                    " match the number of outputs for the reaction"
                    " ({})".format(len(successors), len(reaction.outputs))
                )


def _used_in_designs(ndata: Dict) -> Dict:
    x = sorted(list(ndata["design_keys"]))
    return [{"design_key": _x[0], "assembly": _x[1]} for _x in x]


def _reaction_metadata(reaction):
    return {
        "material_cost": reaction.metadata["material"],
        "efficiency": reaction.metadata.get("efficiency", 0.0),
        "complexity": reaction.metadata.get("complexity", None),
    }


def _assembly_metadata(assembly):
    rmetas = [_reaction_metadata(r) for r in assembly.nonassembly_reactions]
    material = sum([m["material_cost"] for m in rmetas])
    efficiency = functools.reduce(operator.mul, [m["efficiency"] for m in rmetas])
    complexities = [m["complexity"] for m in rmetas if m["complexity"] is not None]

    if complexities:
        max_complexity = max(complexities)
    else:
        max_complexity = 0

    return {
        "material_cost": material,
        "assembly_efficiency": efficiency,
        "max_complexity": max_complexity,
    }


def _reactions_property(
    graph: nx.DiGraph,
    reaction_node_dict: Dict[str, int],
    molecule_node_dict: Dict[str, int],
) -> Dict:
    property_reaction = []
    for n, index in reaction_node_dict.items():
        ndata = graph.nodes[n]
        reaction = ndata["reaction"]
        output_keys = sorted([mhash(m) for m in reaction.outputs])
        input_keys = sorted([mhash(m) for m in reaction.inputs])
        if reaction.name != Reaction.Types.Assembly:
            metadata = _reaction_metadata(reaction)
        else:
            assembly = ndata["assembly"]
            metadata = _assembly_metadata(assembly)

        property_reaction.append(
            {
                "name": ndata["reaction"].name,
                "index": index,
                "used_in_assemblies": _used_in_designs(ndata),
                "inputs": [molecule_node_dict[n] for n in input_keys],
                "outputs": [molecule_node_dict[n] for n in output_keys],
                "metadata": metadata,
            }
        )
    return property_reaction


def _molecules_property(
    graph: nx.DiGraph,
    reaction_node_dict: Dict[str, int],
    molecule_node_dict: Dict[str, int],
) -> Dict:
    property_molecule = []
    for n, index in molecule_node_dict.items():
        ndata = graph.nodes[n]
        mol = ndata["molecule"]
        used_as_inputs = [reaction_node_dict[n] for n in graph.successors(n)]
        used_as_outputs = [reaction_node_dict[n] for n in graph.predecessors(n)]
        property_molecule.append(
            {
                "name": mol.type.name,
                "index": index,
                "sequence": seqrecord_to_json(mol.sequence),
                "used_in_assemblies": _used_in_designs(ndata),
                "used_as_input_to_reactions": used_as_inputs,
                "used_as_output_to_reactions": used_as_outputs,
            }
        )
    return property_molecule


def _design_property(design, reaction_node_dict):
    status = design.status

    def _reaction_summ(r, a_i):

        return {
            "reaction_index": reaction_node_dict[rhash(r, a_i)],
            "metadata": {
                "cost": r.metadata["cost"],
                "materials": r.metadata["cost"],
                "efficiency": r.metadata.get("efficiency", 0.0),
                "complexity": r.metadata.get("complexity", None),
            },
            "outputs": [
                {
                    "start": m.query_region.a,
                    "end": m.query_region.b,
                    "type": m.type.name,
                }
                for m in r.outputs
            ],
        }

    for qk, qstatus in status.items():
        qstatus["sequence"] = seqrecord_to_json(design.seqdb[qk])
        for a_i, adata in enumerate(qstatus["assemblies"]):
            assembly = design.results[qk].assemblies[a_i]
            adata["summary"] = [
                _reaction_summ(r, a_i) for r in assembly.nonassembly_reactions
            ]
            adata["final_assembly_reaction"] = [
                reaction_node_dict[rhash(r, a_i)] for r in assembly.assembly_reactions
            ]

            # TODO: run start
    return status


def dasi_design_to_output_json(design: Union["Design", "LibraryDesign"]):
    graph = dasi_design_to_dag(design)
    reaction_node_dict = {}
    molecule_node_dict = {}
    sorted_nodes = list(nx.topological_sort(graph))[::-1]
    for n in sorted_nodes:
        ndata = graph.nodes[n]
        r = ndata.get("reaction", None)
        m = ndata.get("molecule", None)
        if r:
            ndata["index"] = len(reaction_node_dict)
            assert n not in reaction_node_dict
            reaction_node_dict[n] = ndata["index"]
        elif m:
            ndata["index"] = len(molecule_node_dict)
            assert n not in molecule_node_dict
            molecule_node_dict[n] = ndata["index"]

            if m.type.name == C.PLASMID:
                pass
        else:
            raise ValueError

    return {
        "metadata": design.metadata,
        "designs": _design_property(design, reaction_node_dict),
        "molecules": _molecules_property(graph, reaction_node_dict, molecule_node_dict),
        "reactions": _reactions_property(graph, reaction_node_dict, molecule_node_dict),
    }



import hashlib
import os
from glob import glob
from os.path import abspath
from os.path import dirname
from os.path import isfile
from os.path import join

from dasi.cost import span_cost
from dasi.cost.span_cost import SpanCost
from dasi.log import logger

here = abspath(dirname(__file__))
DEFAULT_COST_FILEPATH = join(here, "span_cost.b")
DEFAULT_COST_CHECKSUM = join(here, "span_cost.checksum")


def hashfiles(files, hash="sha1", hashfunc=None):
    if not hashfunc:

        def hashfunc(string):
            return getattr(hashlib, hash)(string.encode("utf-8")).hexdigest()

    contents = ""
    sorted_path = sorted(files)
    for file in sorted_path:
        with open(file, "r") as f:
            contents += hashfunc(f.read())
    return hashfunc(contents)


def cost_module_checksum():
    cost_dir = dirname(span_cost.__file__)
    cost_files = sorted(glob(join(cost_dir, "*.py")) + glob(join(cost_dir, "*.json")))
    return hashfiles(cost_files)


def cached(path, save_func, load_func, checksum_path, logger=None):
    different_checksum = True
    checksum = cost_module_checksum()
    if os.path.isfile(checksum_path):
        with open(checksum_path, "r") as f:
            stored_checksum = f.read().strip()
            if stored_checksum == checksum:
                different_checksum = False
            if logger:
                logger.debug("Stored checksum: {}".format(stored_checksum))
    if logger:
        logger.debug("Checksum: {}".format(checksum))
    if different_checksum or not isfile(path):
        if logger:
            logger.debug("Using default params")
        model = save_func(path)
        with open(checksum_path, "w") as f:
            f.write(checksum)
        stat = os.stat(path)
        if logger:
            logger.debug("Wrote {} bytes".format(stat.st_size))
    else:
        if logger:
            logger.debug("Loading {}".format(path))
        stat = os.stat(path)
        if logger:
            logger.debug("Loaded {} bytes".format(stat.st_size))
        model = load_func(path)
    return model


def cached_span_cost(
    cost_filepath=DEFAULT_COST_FILEPATH, cost_checksum_filepath=DEFAULT_COST_CHECKSUM
):
    """This will check the checksum of the cost module against the last
    checksum. If checksums are the same, the span cost will be loaded. Else,
    span_cost will be created from default parameters and saved with the cost
    module's checksum.

    :param cost_filepath: path of the span_cost
    :param cost_checksum_filepath: path of the checksum
    :return: SpanCost
    """

    def load_span_cost(path):
        span_cost = SpanCost.load(path)
        return span_cost

    def save_span_cost(path):
        span_cost = SpanCost.open()
        span_cost.dump(path)
        return span_cost

    return cached(
        cost_filepath,
        load_func=load_span_cost,
        save_func=save_span_cost,
        checksum_path=cost_checksum_filepath,
        logger=logger,
    )



"""Cost utilities."""
import numpy as np


def square_broadcast(a, b, a_axis=0, b_axis=1):
    """Broadcast b to the a_axis of ndarray 'a' and perform a hstack."""
    c = np.broadcast_to(b, (a.shape[a_axis], b.shape[b_axis]))
    return np.hstack([a, c])


def df_to_np_ranged(min_col, max_col, df, cols=None, dtype=None):
    """Expand a pandas data frame, which has min and max span columns defined,
    into a numpy ndarray. Specific columns to include in the expansion can be
    defined in the `remaining_cols` argument, else all columns are used.

    For example:

    .. code-block::

      df = pd.DataFrame([
        [0, 10, 10], [10, 20, 100]
      ], columns=['min', 'max', 'x'])
      a = df_to_np_ranged('min', 'max', df, dtype=np.float64)
      assert a.shape == (20, 2)
      assert a[5] == [5., 10.]
      assert a[17] == [17., 100.]
    """
    vblocks = []
    mn, mx = min_col, max_col
    if cols is None:
        cols = [c for c in df.columns if c not in [mn, mx]]
    for _, row in df.iterrows():
        _a = np.arange(row[mn], row[mx], dtype=np.int32).reshape(-1, 1)
        cost_row = row[cols]
        vblocks.append(square_broadcast(_a, cost_row.to_numpy(dtype=dtype), 0, 0))

    a = np.vstack(vblocks)
    return a


def unshape(shape, axis):
    if isinstance(axis, int):
        axis = (axis,)
    return [s for i, s in enumerate(shape) if i not in axis]


def flatten_axis(a, axis=None, copy=True):
    """Flatten the array along the specified axes."""
    if axis is None:
        return a.flatten()
    if isinstance(axis, int):
        axis = (axis,)
    to_axis = tuple(range(len(axis)))
    g = np.prod(unshape(a.shape, axis))
    view = np.moveaxis(a, axis, to_axis).reshape(-1, g)
    if copy:
        return view.copy()
    else:
        return view


def duplicates(b, axis=None):
    """Find duplicates within the specified axes.

    If axis not provided, flatten array and return duplicates.
    """
    b = flatten_axis(b, axis, copy=False)
    s = np.sort(b)
    if axis:
        duplicate_list = []
        for x in b:
            duplicate_list.append(duplicates(x))
        return duplicate_list
    return np.unique(s[:-1][s[1:] == s[:-1]])


def lexargmin(a, axis):
    shape = a[0].shape
    reshaped = tuple([_a.reshape(shape[axis], -1) for _a in a])
    i = np.lexsort(reshaped, axis=1)
    i = i[:, :1]
    i = np.unravel_index(i, unshape(shape, axis))
    imin = (np.arange(shape[axis]), *[np.squeeze(_i) for _i in i])
    return imin


class Slicer:
    def __getitem__(self, item):
        return item


slicer = Slicer()



"""biopython.py.

Helper functions for BioPython
"""
import hashlib
import itertools
import random
import re
import tempfile
from copy import deepcopy
from itertools import chain
from typing import List
from typing import Tuple
from typing import Union
from uuid import uuid4

import networkx as nx
from Bio import Restriction
from Bio import SeqIO
from Bio.Alphabet import generic_dna
from Bio.Seq import Seq
from Bio.SeqFeature import CompoundLocation
from Bio.SeqFeature import ExactPosition
from Bio.SeqFeature import FeatureLocation
from Bio.SeqFeature import SeqFeature
from Bio.SeqRecord import SeqRecord
from primer3plus.design import primer3

from .record_to_json import seqrecord_to_json
from dasi.utils.biopython.sequence import anneal
from dasi.utils.biopython.sequence import random_sequence
from dasi.utils.biopython.sequence import rc
from dasi.utils.region import Span

FWD_COLOR = "ApEinfo_fwdcolor"
REV_COLOR = "ApEinfo_revcolor"
CIRCULAR = "circular"
LINEAR = "lienar"
TOPOLOGY = "topology"
DEFAULT_ANNOTATION_TYPE = "misc_feature"


def sort_cycle(arr, key=None):
    """Sort a cyclic array, maintaining order."""
    if key is None:
        arr_with_i = sorted([(x, i) for i, x in enumerate(arr)])
    else:
        arr_with_i = sorted([(key(x), i) for i, x in enumerate(arr)])
    i = arr_with_i[0][1]
    return arr[i:] + arr[:i]


def random_color():
    """Make a random color."""
    random_number = random.randint(0, 16777215)
    hex_number = str(hex(random_number))[2:]
    if len(hex_number) < 6:
        hex_number = "0" * (6 - len(hex_number)) + hex_number
    return "#" + hex_number


def random_slices(mn: int, mx: int, total_length: int):
    """Yield random slices whose lengths sum to the provided total length.

    :param mn: minimum slice size
    :param mx: maximum slice size
    :param total_length: total length of the slices
    :return:
    """
    n = total_length
    j = 0
    while j < n:
        i = j
        j += random.randint(mn, mx)
        if j >= n:
            j = n
        yield (i, j)


def format_float(a, places=2):
    return "%." + str(places) + "f" % round(a, places)


def get_feature_name(feature: SeqFeature):
    return get_feature_qualifiers(feature, "label")


def set_feature_name(feature: SeqFeature, name: str):
    set_feature_qualifier(feature, "label", name)
    return feature


def set_feature_qualifier(feature: SeqFeature, key: str, value):
    feature.qualifiers[key] = [value]


def get_feature_qualifiers(feature: SeqFeature, key: str):
    return feature.qualifiers.get(key, list())[0]


def set_feature_color(feature: SeqFeature, color: str = None, rev_color: str = None):
    if color is None:
        color = random_color()

    if not rev_color:
        rev_color = random_color()

    set_feature_qualifier(feature, FWD_COLOR, color)
    set_feature_qualifier(feature, REV_COLOR, rev_color)


def new_location(i: int, j: int, strand: int) -> FeatureLocation:
    return FeatureLocation(ExactPosition(i), ExactPosition(j), strand=strand)


def new_compound_location(
    indices: List[Union[Tuple[int, int], Tuple[int, int, int]]], strand: int
) -> CompoundLocation:
    locations = []
    for index in indices:
        if not isinstance(index, Tuple):
            raise ValueError(
                "Expects a tuple of integers size 2 or 3, not a {}".format(
                    indices.__class__
                )
            )
        if not len(index) in [2, 3]:
            raise ValueError("Expects a tuple of integers of size 2 or 3")
        if len(index) == 2:
            i, j = index
            s = strand
        elif len(index) == 3:
            i, j, s = index
        else:
            raise ValueError("Must be tuple of 2 or 3 integers")
        if not isinstance(i, int) or not isinstance(j, int) or not isinstance(s, int):
            raise ValueError(
                "Expects a tuple of integers of size 2 or 3. Found {}".format(index)
            )
        locations.append(FeatureLocation(ExactPosition(i), ExactPosition(j), strand=s))
    return CompoundLocation(locations)


def clean_features(record: SeqRecord):
    """Remove redundant features."""
    features_by_hash = {}
    for f in record.features:
        fhash = "*".join([str(f.qualifiers["label"]), f.type, str(f.location)])
        features_by_hash[fhash] = f
    record.features = list(features_by_hash.values())
    return record


def new_compound_feature(
    name: str,
    indices: List[Union[Tuple[int, int], Tuple[int, int, int]]],
    strand: int,
    feature_type: str = None,
):
    location = new_compound_location(indices, strand=strand)
    return new_feature(name, location=location, feature_type=feature_type)


def new_feature(
    name: str,
    location: Union[None, CompoundLocation, FeatureLocation],
    color: str = None,
    feature_type: str = None,
):
    feature = SeqFeature(location=location, qualifiers={}, type=feature_type)
    set_feature_name(feature, name)
    set_feature_color(feature, color=color)
    return feature


def _annotate_feature(
    length: int,
    name: str,
    i: int = None,
    j: int = None,
    cyclic: bool = False,
    feature_type: str = None,
):
    if i is None:
        i = 0
    if j is None:
        j = length

    if cyclic and (j > length or (i > j)):
        if j > length:
            j = j - length
        feature = new_compound_feature(
            name=name,
            indices=[(i, length), (0, j)],
            strand=1,
            feature_type=feature_type,
        )
    else:
        feature = new_feature(
            name=name,
            location=FeatureLocation(ExactPosition(i), ExactPosition(j), strand=1),
            feature_type=feature_type,
        )
    return feature


def annotate(
    record: SeqRecord,
    name: str,
    i: int = None,
    j: int = None,
    cyclic: bool = False,
    annotation_type: str = None,
):
    """Annotate a SeqRecord."""
    if not name:
        raise ValueError("Cannot annotate record with no name.")
    if annotation_type is None:
        annotation_type = DEFAULT_ANNOTATION_TYPE
    feature = _annotate_feature(
        len(record.seq), name, i, j, cyclic, feature_type=annotation_type
    )
    record.features.append(feature)
    return feature


def set_topology(records: List[SeqRecord], topology: str):
    assert topology in [CIRCULAR, LINEAR]
    for r in records:
        r.annotations[TOPOLOGY] = topology


def make_cyclic(records: List[SeqRecord]):
    if isinstance(records, SeqRecord):
        records = [records]
    for r in records:
        r.annotations[TOPOLOGY] = CIRCULAR


def make_linear(records: List[SeqRecord]):
    if isinstance(records, SeqRecord):
        records = [records]
    for r in records:
        r.annotations[TOPOLOGY] = LINEAR


def is_circular(record: SeqRecord):
    return record.annotations.get(TOPOLOGY, False) == CIRCULAR


def is_linear(record: SeqRecord):
    return not is_circular(record)


def new_sequence(
    seqstr: str,
    name: str = None,
    auto_annotate: bool = False,
    cyclic: bool = False,
    annotation_type: str = None,
):
    record = SeqRecord(Seq(seqstr))
    if auto_annotate:
        annotate(record, name, annotation_type=annotation_type)
    if name:
        record.name = name
        record.id = name
        record.description = name

    if cyclic:
        make_cyclic([record])
    else:
        make_linear([record])
    return record


# def slice_feature(feature: SeqFeature, i: int = None, j: int = None):
#     if isinstance(feature)


def _format_slice(slc: slice):
    return "[{}:{}]".format(slc.start, slc.stop)


def slice_with_features(seq: SeqRecord, slc: slice):
    i = slc.start
    j = slc.stop
    if slc.step:
        raise ValueError("Step is not supported.")
    # new_features = []
    # for feature in seq.features:
    #     name = get_feature_name(feature)
    #     new_parts = []
    #     for part in feature.location.parts:
    #         if i in part and j in part:
    #             new_parts.append(new_location(i, j, part.strand))
    #         elif i in part:
    #             new_parts.append(new_location(i, part.end.position, part.strand))
    #         elif j in part:
    #             new_parts.append(new_location(part.start.position, j, part.strand))
    #     if len(new_parts) == 1:
    #         new_feature = deepcopy(feature)
    #         new_feature.location = new_parts[0]
    #         set_feature_name(new_feature, name + _format_slice(slc))
    #         new_features.append(new_feature)
    #     elif len(new_parts) > 2:
    #         new_feature = deepcopy(feature)
    #         new_feature.location = CompoundLocation(new_parts)
    #         set_feature_name(new_feature, name + _format_slice(slc))
    #         new_features.append(new_feature)
    new_seq = seq[i:j]
    # new_seq.features += new_features
    return new_seq


def random_record(
    length: int, name: str = None, auto_annotate: bool = True, cyclic: bool = False
):
    if name is None:
        name = str(uuid4())
    return new_sequence(
        random_sequence(length), name=name, auto_annotate=auto_annotate, cyclic=cyclic
    )


def randomly_annotate(
    rec: SeqRecord,
    feature_length_range: Tuple[int, int],
    feature_name_list: List[str] = None,
) -> SeqRecord:
    for i, j in random_slices(
        feature_length_range[0], feature_length_range[1], len(rec.seq)
    ):
        if not feature_name_list:
            random_name = str(uuid4())[-4:]
        else:
            random_name = random.sample(feature_name_list, k=1)[0]
        annotate(rec, random_name, i, j)
    return rec


def pcr_amplify(
    fwd: SeqRecord,
    rev: SeqRecord,
    template: SeqRecord,
    cyclic: bool,
    cyclic_buffer: int = 100,
    name: str = None,
    return_matches: bool = False,
):
    original_template = template
    if cyclic:
        template = template + template + template[:cyclic_buffer]
    fwd_matches, rev_matches = anneal(str(template.seq), [str(fwd.seq), str(rev.seq)])

    products_by_sequence = {}
    for f, r in itertools.product(fwd_matches, rev_matches):
        i = f["top_strand_slice"][0]
        j = r["top_strand_slice"][1]

        try:
            span = Span(
                i,
                j,
                context_length=len(original_template.seq),
                cyclic=cyclic,
                ignore_wrap=True,
            )
        except IndexError as e:
            if not cyclic:
                continue
            else:
                raise e

        f_rec = new_sequence(
            f["overhang"], name=fwd.name + "_overhang", auto_annotate=True
        )
        r_rec = new_sequence(
            rc(r["overhang"]), name=rev.name + "_overhang", auto_annotate=True
        )

        product = span.get_slice(original_template)
        # annotate(template_chunk, name="source: {}".format(template_name[:40]))
        if len(f_rec.seq):
            product = f_rec + product
        if len(r_rec.seq):
            product = product + r_rec

        if not name:
            name = original_template.name or original_template.id
            name += "[{}:{}]".format(span.a, span.b)
        # annotate(product, name=name)
        if len(product) <= len(original_template.seq):
            products_by_sequence[str(product.seq)] = (product, span.a, span.b)
    product_list = list(products_by_sequence.values())
    product_list.sort(key=lambda x: (x[1], len(product)))
    if return_matches:
        return product_list, fwd_matches, rev_matches
    return product_list


def load_glob(*paths: Tuple[str, ...], format: str = None):
    path_iter = chain(*paths)
    records = []
    for path in path_iter:
        records += SeqIO.parse(path, format=format)
    return records


def load_fasta_glob(path: str) -> List[SeqRecord]:
    return load_glob(path, format="fasta")


def load_genbank_glob(path: str) -> List[SeqRecord]:
    return load_glob(path, format="genbank")


def write_tmp_records(records: List[SeqRecord], format: str) -> List[SeqRecord]:
    fd, tmp_path_handle = tempfile.mkstemp(suffix="." + format)
    SeqIO.write(records, tmp_path_handle, format=format)
    return tmp_path_handle


class GibsonAssembler:
    @staticmethod
    def make_hash(s: str):
        return hashlib.sha1(s.encode("utf-8")).hexdigest()

    @classmethod
    def rnode(cls, record: SeqRecord):
        return cls.make_hash(str(record.name + record.seq))

    @classmethod
    def add_edge(
        cls, g: nx.DiGraph, r1: SeqRecord, r2: SeqRecord, tm: float, match: dict
    ):
        n1 = cls.rnode(r1)
        n2 = cls.rnode(r2)
        if n1 not in g:
            g.add_node(n1, record=r1)
        if n2 not in g:
            g.add_node(n2, record=r2)
        g.add_edge(n1, n2, tm=tm, match=match)

    @classmethod
    def interaction_graph(cls, records: List[SeqRecord]) -> nx.DiGraph:
        g = nx.DiGraph()
        pairs = []
        for r1, r2 in itertools.product(records, records):
            if r1 is not r2:
                pairs.append((r1, r2))
                if r2 is not records[0]:
                    pairs.append((r1, r2.reverse_complement(name=r2.name + "_rc")))

        for r1, r2 in pairs:
            fwd, rev = anneal(str(r1.seq), [str(r2.seq)])
            for f in fwd:
                if f["top_strand_slice"][0] == 0:
                    anneal_seq = f["anneal"]
                    tm = primer3.calcTm(anneal_seq[-60:])
                    cls.add_edge(g, r2, r1, tm=tm, match=f)
        return g

    @classmethod
    def _collect_features(cls, records: List[SeqRecord]):
        seq_to_features = {}
        for record in records:
            for feature in record.features:
                s = str(feature.extract(record.seq)).upper()
                if s not in seq_to_features:
                    seq_to_features[s] = feature
        return seq_to_features

    @classmethod
    def _restore_features(cls, seq_to_features, record, cyclic: bool):
        for seq in seq_to_features:
            record_seq = str(record.seq)
            if cyclic:
                record_seq += record_seq
            matches = list(re.finditer(seq, record_seq, re.IGNORECASE))
            if len(matches) >= 1:
                feature = seq_to_features[seq]
                new_feature = deepcopy(feature)
                for match in matches:
                    span = match.span()

                    if span[0] < len(record.seq):
                        if span[1] >= len(record.seq):
                            f = _annotate_feature(
                                len(record.seq), "", span[0], span[1], cyclic=True
                            )
                            new_feature.location = f.location
                        else:
                            new_feature.location = new_location(
                                span[0], span[1], feature.location.strand
                            )
                    record.features.append(new_feature)

    @classmethod
    def assemble_record_from_cycle(
        cls,
        cycle: List[str],
        graph: nx.DiGraph,
        annotate_sources: bool = False,
        annotate_junctions: bool = False,
    ) -> SeqRecord:

        source_indices = []
        i = 0

        records = [graph.nodes[n]["record"] for n in cycle]
        stored_features = cls._collect_features(records)

        record = SeqRecord(Seq("", alphabet=generic_dna))

        c1 = cycle[-1:] + cycle[:-1]
        c2 = cycle
        c3 = cycle[1:] + cycle[:1]

        for n1, n2, n3 in zip(c1, c2, c3):
            r1, r2, r3 = [graph.nodes[n]["record"] for n in [n1, n2, n3]]

            edata12 = graph.get_edge_data(n1, n2)
            edata23 = graph.get_edge_data(n2, n3)

            s12 = edata12["match"]["top_strand_slice"]
            s23 = edata23["match"]["top_strand_slice"]

            o12 = r2[s12[0] : s12[1]]
            o23 = r3[s23[0] : s23[1]]

            trunc_r2 = slice_with_features(r2, slice(len(o12), -len(o23)))

            if annotate_junctions:
                annotate(o12, "junction: tm {}".format(format_float(edata12["tm"])))
            record += o12
            record += trunc_r2

            source_indices.append((i, i + len(r2)))
            i = i + len(r2) - len(o23)

        for i, (a, b) in enumerate(source_indices):
            if annotate_sources:
                annotate(
                    record,
                    "source: fragment {} ({})".format(i, r2.name),
                    a,
                    b,
                    cyclic=True,
                )

        cls._restore_features(stored_features, record, cyclic=True)
        clean_features(record)
        return record

    @classmethod
    def make_cyclic_assemblies(
        cls,
        records: List[SeqRecord],
        annotate_sources: bool = False,
        annotate_junctions: bool = False,
    ) -> List[SeqRecord]:
        g = cls.interaction_graph(records)

        node_rank = {}
        for i, r in enumerate(records):
            node_rank[cls.rnode(r)] = i

        cycles = list(nx.simple_cycles(g))
        cycles = [sort_cycle(cycle, key=lambda n: node_rank[n]) for cycle in cycles]

        assembled_records = []
        for cycle in cycles:
            record = cls.assemble_record_from_cycle(
                cycle,
                g,
                annotate_sources=annotate_sources,
                annotate_junctions=annotate_junctions,
            )
            assembled_records.append(record)
        make_cyclic(assembled_records)
        return assembled_records


def digest(record: SeqRecord, restriction_enzyme: str):
    """Digest a SeqRecord with an enzyme.

    :param record: record to digest
    :param restriction_enzyme: name of restriction enzyme
    :return: list of records
    """
    enzyme = getattr(Restriction, restriction_enzyme)
    linear = is_linear(record)
    indices = enzyme.search(record.seq, linear=linear)
    indices = [i - 1 for i in indices]
    pairs = zip(indices[:-1], indices[1:])
    records = []
    for i1, i2 in pairs:
        records.append(record[i1:i2])
    if not linear:
        records.append(record[indices[-1] :] + record[: indices[0]])
    return records


def random_record_from_library(
    records: List[SeqRecord],
    circular: bool,
    size_interval: Tuple[int, int] = (5000, 10000),
    max_chunks: int = None,
    chunk_size_interval: Tuple[int, int] = (100, 3000),
    random_chunk_prob_int: Tuple[float, float] = (0, 0.5),
    random_chunk_size_int: Tuple[int, int] = (100, 1000),
):
    """

    :param records:
    :param size_interval:
    :param chunk_size_interval:
    :param random_chunk_prob_interval: picks a random probability at whic
    :param circular:
    :return:
    """
    length = random.randint(*size_interval)
    new_record = new_sequence("")

    x = random.uniform(*random_chunk_prob_int)

    n_chunks = 0
    while len(new_record.seq) < length and (
        max_chunks is None or n_chunks < max_chunks
    ):
        if random.random() < x:
            rand_chunk_len = random.randint(*random_chunk_size_int)
            rand_rec = random_record(rand_chunk_len)
        else:
            rand_chunk_len = random.randint(*chunk_size_interval)
            rand_rec = random.choice(records)
            if rand_chunk_len > len(rand_rec.seq):
                continue
                # rand_chunk_len = len(rand_rec.seq)

        i = random.randint(0, len(rand_rec.seq) - rand_chunk_len)
        rand_chunk = rand_rec[i : i + rand_chunk_len]
        if random.random() < 0.5:
            rand_chunk = rand_chunk.reverse_complement()

        new_record += rand_chunk
        n_chunks += 1
    new_record = new_record[:length]
    if circular:
        make_cyclic([new_record])
    else:
        make_linear([new_record])
    return new_record


make_cyclic_assemblies = GibsonAssembler.make_cyclic_assemblies



import json
from abc import ABC
from abc import abstractmethod
from functools import partial
from os.path import abspath
from os.path import dirname
from os.path import join
from typing import Tuple
from typing import Union

import jsonschema
import msgpack
import numpy as np
import pandas as pd
import seaborn as sns

from .utils import df_to_np_ranged
from .utils import lexargmin
from .utils import slicer
from dasi.exceptions import DasiCostParameterValidationError
from dasi.schemas import Schemas
from dasi.schemas import validate_with_schema
from dasi.utils import NumpyDataFrame

here = abspath(dirname(__file__))


def encoder(obj):
    """msgpack encoder for cost functions.

    :param obj:
    :return:
    """
    if isinstance(obj, NumpyDataFrame):
        return {"__numpydataframe__": True, "data": obj.data}
    elif isinstance(obj, SpanCost):
        return {
            "__spancost__": True,
            "cost_dict": {k: v for k, v in obj.cost_dict.items()},
            "span": obj.span,
        }
    elif isinstance(obj, PrimerCostModel):
        return (
            {
                "__primercostmodel__": True,
                "cost_dict": {k: v for k, v in obj.cost_dict.items()},
                "span": obj.span,
            },
        )
    elif isinstance(obj, SynthesisCostModel):
        return {
            "__synthesiscostmodel__": True,
            "cost_dict": {k: v for k, v in obj.cost_dict.items()},
            "span": obj.span,
        }
    return obj


def decoder(obj):
    """msgpack decoder for cost functions.

    :param obj:
    :return:
    """
    if b"__numpydataframe__" in obj:
        data = obj[b"data"]
        data = {k.decode(): v for k, v in data.items()}
        obj = NumpyDataFrame(data=data)
    elif b"__primercostmodel__" in obj:
        cost_dict = {tuple(k): v for k, v in obj[b"cost_dict"].items()}
        span = obj[b"span"]
        obj = PrimerCostModel.__new__(PrimerCostModel)
        obj.cost_dict = cost_dict
        obj.span = span
    elif b"__synthesiscostmodel__" in obj:
        cost_dict = {tuple(k): v for k, v in obj[b"cost_dict"].items()}
        span = obj[b"span"]
        obj = SynthesisCostModel.__new__(SynthesisCostModel)
        obj.cost_dict = cost_dict
        obj.span = span
    elif b"__spancost__" in obj:
        cost_dict = {tuple(k): v for k, v in obj[b"cost_dict"].items()}
        span = obj[b"span"]
        obj = SpanCost.__new__(SpanCost)
        obj.cost_dict = cost_dict
        obj.span = span
    return obj


def _add_special_primer_case(params):
    primer_cost_data = params["primer_cost"]["data"]
    primer_cost_columns = params["primer_cost"]["columns"]
    i = primer_cost_columns.index("min")
    j = primer_cost_columns.index("max")
    k = primer_cost_columns.index("name")

    new = [0] * len(primer_cost_columns)

    new[k] = "Non-primer (special case)"
    new[i] = params["primer_min_anneal"]
    new[j] = params["primer_min_anneal"] + 1
    if new not in primer_cost_data:
        primer_cost_data.insert(0, new)


def _replace_inf(params):
    if isinstance(params, list):
        for i, p in enumerate(params):
            if p == "inf":
                params[i] = np.inf
        for p in params:
            _replace_inf(p)
    elif isinstance(params, dict):

        for k, v in params.items():
            if v == "inf":
                params[k] = np.inf
            else:
                _replace_inf(v)


def validate_params(params):
    schema = Schemas.cost_parameters_schema
    validate_with_schema(params, schema, reraise_as=DasiCostParameterValidationError)


default_parameters_path = join(here, "default_parameters.json")


def open_params(path=None):
    if path is None:
        path = default_parameters_path
    with open(path, "r") as f:
        params = json.load(f)
    return params


class CostBuilder(ABC):
    """Abstract base class for building a cost model."""

    def __init__(self, span: np.ndarray):
        """Initialize cost builder from a 'span', or a list of indices for
        which the model is valid.

        :param span: the list of indices (span) for which this cost model is evaluated.
        """
        self.cost_dict = {}
        self.span = span

    @staticmethod
    def _fix_params(params):
        validate_params(params)
        _add_special_primer_case(params)
        _replace_inf(params)
        return params

    @classmethod
    @abstractmethod
    def from_json(cls, params: dict, override_span: np.ndarray = None):
        """Initialize from parameters."""
        cls._fix_params(params)

    @classmethod
    def open(cls, path: str = None):
        params = open_params(path)
        return cls.from_json(params)

    @abstractmethod
    def compute(self):
        """Before the numpy-based cost calculations."""
        pass

    def to_df(self):
        """Convert the cost builder to a pandas.DataFrame."""
        dfs = []
        for ext, jxn in self.cost_dict.items():
            df = jxn.to_df()
            df["condition"] = [ext] * len(df)
            dfs.append(df)
        df = pd.concat(dfs)
        return df

    @property
    def plot(self):
        """Plot attributes across 'span' from the pandas.DataFrame."""
        return partial(
            sns.lineplot, data=self.to_df(), x="span", y="cost", hue="condition"
        )

    # TODO: what if there is a gap in the span?
    def cost(
        self, bp: Union[np.ndarray, int], ext: Tuple[int, int], invalidate=False
    ) -> NumpyDataFrame:
        """Return the :class:`NumpyDataFrame <dasi.usil.NumpyDataFrame>` across
        the provided span.

        :param bp: either a integer or np.ndarray of integers to compute.
        :param ext: The extension design parameters (bool, bool) that represent
                    whether the left or right primer is 'extendable'. Primers
                    that have been provided are not extendable. If we had,
                    for example, an existing right primer and flexibility to
                    design the left primer, this would be `(1, 0)`.
        :param invalidate: Whether to invalidate indices that go beyond the provided
                            span for the cost builder (`CostModelBase.span`). Any
                            invalid indices cost will be set to `np.inf` and efficiency
                            to `0.0`
        :return: NumpyDataFrame representing the cost of from the input span.
        """
        if isinstance(bp, int):
            bp = np.array([bp])
        _span = self.span.flatten()
        index = np.array(bp - _span.min(), dtype=np.int64)

        # clipped index
        clipped = np.clip(index, 0, len(_span) - 1)
        if isinstance(clipped, int):
            index = np.array([clipped])

        # get junction
        jxn = self.cost_dict[ext][clipped]

        if invalidate:
            # invalidated indices
            invalid_i1 = index < 0
            invalid_i2 = index >= len(_span)

            # invalidate extremes
            for invalid in [invalid_i1, invalid_i2]:
                jxn.data["cost"][invalid] = np.inf
                jxn.data["efficiency"][invalid] = 0.0

        # add span
        jxn.col["span"] = np.array(bp)
        return jxn

    def __call__(
        self, bp: Union[int, np.ndarray], ext: Tuple[int, int]
    ) -> NumpyDataFrame:
        """Short hand for calling :meth:`cost`"""
        return self.cost(bp, ext)


class PrimerCostModel(CostBuilder):
    def __init__(
        self,
        pdf: pd.DataFrame,
        edf: pd.DataFrame,
        min_anneal: int,
        time_cost: float,
        material_mod: float,
        min_span: int,
        input_data: dict = None,
        override_span: np.ndarray = None,
    ):
        """Initialize the cost model for primer designs.

        :param pdf: a pandas DataFrame representing costs for primers.
                    Example parameters found at
                    :attr:`dasi.cost.params.PrimerParams.primer_df`
        :param edf: a pandas DataFrame representing efficiencies for spans.
                    Example parameters found at
                    :attr:`dasi.cost.params.PrimerParams.eff_df`
        :param min_anneal: the number of bases to consider for annealing a primer.
        :param time_cost: the cost to produce a primer
        :param material_mod:
        :param min_span:
        """
        self.input_data = input_data
        self.primer_df = pdf
        self.eff_df = edf
        self.time_cost = time_cost
        self.min_anneal = min_anneal
        max_span = self.primer_df["max"].max() * 2 - min_span
        if override_span is not None:
            span = override_span
        else:
            span = np.arange(min_span, max_span, dtype=np.int32)
        super().__init__(span)
        self.material_modifier = material_mod
        self.compute()

    @classmethod
    def from_json(cls, params, override_span: np.ndarray = None) -> "PrimerCostModel":
        """Load from :class:`dasi.cost.params.PrimerParams`.

        :param primer_params: parameters
        :return: PrimerCostModel
        """
        super().from_json(params)
        return PrimerCostModel(
            pdf=pd.DataFrame(
                params["primer_cost"]["data"], columns=params["primer_cost"]["columns"]
            ),
            edf=pd.DataFrame(
                params["primer_efficiency"]["data"],
                columns=params["primer_efficiency"]["columns"],
            ),
            time_cost=params["global_time_cost"],
            material_mod=params["global_material_modifier"],
            min_span=params["_primer_min_span"],
            min_anneal=params["primer_min_anneal"],
            input_data=params,
            override_span=override_span,
        )

    # TODO: IMPORTANT!! For extreme values, for extensions material cost should NOT
    #       BE ZERO
    def compute(self):
        span = self.span

        # span, base cost, cost per bp, time (days)
        p = df_to_np_ranged(
            "min",
            "max",
            self.primer_df,
            cols=["base cost", "cost per bp", "time (days)"],
            dtype=np.float64,
        )

        # flattened extension array
        ext = p[:, 0].reshape(-1, 1) - self.min_anneal
        ext = ext.astype(np.int32)

        # relative span (i.e. the overlap)
        rel_span = span[:, np.newaxis, np.newaxis] - (ext + ext.T)[np.newaxis, :, :]

        # efficiency, the same shape as rel_span
        eff_arr = df_to_np_ranged("min", "max", self.eff_df, dtype=np.float64)[:, 1]
        eff = eff_arr[np.clip(-rel_span, 0, len(eff_arr) - 1)]

        # material cost
        m = p[:, 0, np.newaxis] * p[:, 2, np.newaxis] + p[:, 1, np.newaxis]
        t = p[:, 3, np.newaxis]
        t = np.maximum(t, t.T)
        x = m * self.material_modifier + t * self.time_cost
        material_cost = x + x.T

        # cost
        cost = material_cost / eff
        cost[np.where(np.isnan(cost))] = np.inf

        slice_dict = {
            (0, 0): slicer[:, :1, :1],
            (0, 1): slicer[:, :1, 1:],
            (1, 0): slicer[:, 1:, :1],
            (1, 1): slicer[:, 1:, 1:],
        }

        for slice_index, slice_obj in slice_dict.items():
            s_eff = eff[slice_obj]
            s_cost = cost[slice_obj]
            s_mat = material_cost[slice_obj[1], slice_obj[2]]

            idx = lexargmin((s_eff, s_cost), axis=0)
            self.cost_dict[slice_index] = NumpyDataFrame(
                dict(
                    span=span[idx[0]],
                    cost=s_cost[idx],
                    efficiency=s_eff[idx],
                    material=s_mat[idx[1], idx[2]],
                    left_ext=ext[idx[1]],
                    right_ext=ext[idx[2]],
                    time=t[idx[1], idx[2]],
                ),
                apply=np.squeeze,
            )


class SynthesisCostModel(CostBuilder):
    def __init__(
        self,
        sdf: pd.DataFrame,
        primer_cost: PrimerCostModel,
        time_cost: float,
        material_modifier: float,
        step_size=10,
        left_span_range=(-500, 500),
        input_data: dict = None,
        override_span: np.ndarray = None,
    ):
        """

        :param sdf: synthesis parameter pandas dataframe. Example parameters found
                    at :attr:`dasi.cost.params.SynthesisParams.synthesis_df`
        :param primer_cost: primer cost builder
        :param time_cost: time cost per day
        :param material_modifier: material multiplier (default: 1.0)
        :param step_size: step size to consider (default: 10)
        :param left_span_range: span range for left primer to consider
                (default: -500, 500)
        :param input_data: optional input data used to created the cost model
        :param override_span: optionally, only compute over the provided span (for
                debugging/testing)
        """
        self.input_data = input_data
        self.synthesis_df = sdf
        self.step_size = step_size
        self.material_modifier = (material_modifier,)
        self.lspanrange = left_span_range
        self.primer_cost = primer_cost
        self.time_cost = time_cost
        self.syn = df_to_np_ranged("min", "max", self.synthesis_df, dtype=np.float32)
        if override_span is not None:
            span = override_span
        else:
            span = self.syn[:, 0].reshape(1, -1).astype(np.int32)
        super().__init__(span)
        self.compute()

    @classmethod
    def from_json(
        cls,
        params: dict,
        primer_cost_model: PrimerCostModel,
        override_span: np.ndarray = None,
    ):
        super().from_json(params)
        return SynthesisCostModel(
            sdf=pd.DataFrame(
                params["synthesis_cost"]["data"],
                columns=params["synthesis_cost"]["columns"],
            ),
            primer_cost=primer_cost_model,
            time_cost=params["global_time_cost"],
            material_modifier=params["global_material_modifier"],
            step_size=params["_synthesis_step_size"],
            left_span_range=params["_synthesis_left_span_range"],
            override_span=override_span,
            input_data=params,
        )

    @classmethod
    def open(cls, path: str = None, primer_cost: PrimerCostModel = None):
        params = open_params(path)
        if not primer_cost:
            primer_cost = PrimerCostModel.open(path)
        return cls.from_json(params, primer_cost)

    def compute(self):
        gene_sizes = (
            self.syn[:, 0].reshape(-1, 1)[:: self.step_size, :].astype(np.int32)
        )
        gene_costs = self.syn[:, 1][gene_sizes]
        gene_times = self.syn[:, 2][gene_sizes]

        left_span = np.arange(self.lspanrange[0], self.lspanrange[1], self.step_size)[
            ..., np.newaxis, np.newaxis
        ]

        for i, j in [(0, 0), (1, 0), (0, 1), (1, 1)]:
            jxn = self._compute(gene_costs, gene_sizes, gene_times, i, j, left_span)
            self.cost_dict[(i, j)] = jxn

    def _compute(
        self,
        gene_costs,
        gene_sizes,
        gene_times,
        i: Union[bool, int],
        j: Union[bool, int],
        left_span,
    ):
        # extension conditions, idk
        left_ext = (i, 0)
        right_ext = (0, j)
        # left primer
        left_jxn = self.primer_cost(left_span, ext=left_ext)
        left_eff = left_jxn.data["efficiency"]
        left_material = left_jxn.data["material"]

        # right primer
        right_span = self.span - gene_sizes - left_span
        right_jxn = self.primer_cost(right_span, ext=right_ext)
        right_eff = right_jxn.data["efficiency"]
        right_material = right_jxn.data["material"]
        ext_material = left_material + right_material
        ext_eff = np.multiply(left_eff, right_eff)

        # swap axes
        # span, size, left_span
        ext_material = ext_material.swapaxes(0, 2)
        ext_eff = ext_eff.swapaxes(0, 2)
        syn_eff = ext_eff * 1.0  # here place probability of success for gene synthesis
        # could even use sequence to compute this later???
        syn_material_cost = (
            ext_material + gene_costs[np.newaxis, ...] * self.material_modifier
        )
        syn_time_cost = gene_times * self.time_cost
        syn_total_cost = (syn_material_cost + syn_time_cost[np.newaxis, ...]) / syn_eff
        idx = lexargmin((syn_eff, syn_total_cost), axis=0)

        _gcosts = gene_costs[idx[1]]
        _span = np.squeeze(self.span)[idx[0]]
        _gtimes = syn_time_cost[idx[1]]
        gene_df = NumpyDataFrame(
            dict(
                cost=_gcosts,
                material=_gcosts,
                time=_gtimes,
                efficiency=np.ones(idx[0].shape[0]),
                size=gene_sizes[idx[1]],
            ),
            apply=np.squeeze,
        )

        flat_left_jxn = left_jxn[idx[2]].apply(np.squeeze)
        flat_right_jxn = right_jxn[idx[2], idx[1], idx[0]]

        time = np.vstack(
            (
                flat_left_jxn.data["time"],
                flat_right_jxn.data["time"],
                gene_df.data["time"],
            )
        ).max(axis=0)

        gap_df = NumpyDataFrame(
            dict(
                span=_span,
                cost=syn_total_cost[idx],
                efficiency=syn_eff[idx],
                time=time,
                material=syn_material_cost[idx],
                lshift=left_span[idx[2]],
            ),
            apply=np.squeeze,
        )

        gap_df.update(flat_left_jxn.prefix("lprimer_"))
        gap_df.update(flat_right_jxn.prefix("rprimer_"))
        gap_df.update(gene_df.prefix("gene_"))

        return gap_df

    def cost(
        self, bp: Union[int, np.ndarray], ext: Tuple[int, int], invalidate=True
    ) -> NumpyDataFrame:
        return super().cost(bp, ext, invalidate=invalidate)

    def __call__(self, bp, ext):
        return self.cost(bp, ext)


class SpanCost(CostBuilder):
    def __init__(
        self, syn_cost, input_data: dict = None, override_span: np.ndarray = None
    ):
        self.syn_cost = syn_cost
        self.primer_cost = self.syn_cost.primer_cost
        x = [
            syn_cost.span.min(),
            syn_cost.span.max(),
            self.primer_cost.span.min(),
            self.primer_cost.span.max(),
        ]
        if override_span is not None:
            span = override_span
        else:
            span = np.arange(min(x), max(x))
        self.input_data = input_data
        super().__init__(span)
        self.compute()

    @classmethod
    def from_json(cls, params, override_span: np.ndarray = None):
        super().from_json(params)
        primer_cost_model = PrimerCostModel.from_json(params)
        synthesis_cost_model = SynthesisCostModel.from_json(
            params, primer_cost_model, override_span=override_span
        )
        return cls(
            syn_cost=synthesis_cost_model,
            input_data=params,
            override_span=override_span,
        )

    def compute(self):
        def choose(a, i):
            return np.choose(i, a)

        for ext in [(0, 0), (0, 1), (1, 0), (1, 1)]:
            # numpy data frames for primer cost and syn cost over span
            df1 = self.primer_cost(self.span, ext)
            df2 = self.syn_cost(self.span, ext)

            # determine the indices of the min cost (0=primer, 1=syn)
            c1 = df1.data["cost"]
            c2 = df2.data["cost"]
            c3 = np.stack((c1, c2), axis=1)
            y = c3.argmin(axis=1)

            # select between primer_cost and syn_cost based on the min cost
            df4 = NumpyDataFrame.group_apply(
                (df1, df2), choose, i=y, _fill_value=np.nan
            )
            self.cost_dict[ext] = df4

    def cost(
        self, bp: Union[int, np.ndarray], ext: Tuple[int, int], invalidate=True
    ) -> NumpyDataFrame:
        return super().cost(bp, ext, invalidate=invalidate)

    def dumpb(self) -> bytes:
        return msgpack.packb(self, default=encoder, use_bin_type=True)

    @classmethod
    def loadb(cls, s: bytes):
        return msgpack.unpackb(s, object_hook=decoder, raw=True, use_list=False)

    def dump(self, path: str):
        with open(path, "wb") as f:
            f.write(self.dumpb())

    @classmethod
    def load(cls, path: str):
        with open(path, "rb") as f:
            return cls.loadb(f.read())



"""NumpyDataFrame."""
import pprint
from collections import OrderedDict
from collections.abc import Mapping as MappingABC
from typing import Any
from typing import Callable
from typing import Dict
from typing import Generator
from typing import ItemsView
from typing import Iterable
from typing import Iterable as IterableABC
from typing import List
from typing import Tuple
from typing import Union

import msgpack
import msgpack_numpy as m
import numpy as np
import pandas as pd

m.patch()


class Null:
    """Not None."""


class NumpyDataFrameException(Exception):
    """Generic exceptions for NumpyDataFrame."""


class NumpyDataFrame(MappingABC):
    """The NumpyDataFrame is a class halfway between pandas and numpy. It has
    named columns, indexing, slicing, function applications, and mathematical
    operations. Unlike pandas however, it maintains the multi-dimensionality of
    underlying data (as np.ndarray), allowing broadcasting and complex
    indexing.

    Usage:

    **indexing and columns**

    All of the underlying arrays can be slices and indexed using the slicing operations.
    The native np.ndarray indexing is used for each column, meaning a *view* is returned
    with the same memory locations.

    .. code-block::

        df = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
        df[0]
        df[0, 1]
        df[np.array([0, 1])]

    Columns can be selected and returned:

    .. code-block::

        df.col['A', 'B']    # return df with 'A' and 'B'
        df.col['A']         # return df with only 'A'
        print(list(df.col)) # return the column names
        print(df.columns)   # also returns the column names

    New columns can be added:

    .. code-block::

        df.col['A'] = np.arange(10)

    Columns can be deleted:

    .. code-block::

        del df.col['B']

    Add prefix or suffix to column names:

      df.prefix('prefix_', cols=['A']) # add prefix only to 'A', return new df
      df.suffix('__suffix')             # add suffix to all columns, return new df

    **apply**

    Functions can be apply to each column using `np.apply`.
    For example, the following applies `np.reshape` to all of the columns individually,
    returning a new df:

    .. code-block::

      df = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
      df.apply(np.reshape, (-1, 1))

    Functions can be applied to tuples of all of the columns by `np.aggregate`.
    For example, the following stacks all of the columns horizontally, returning
    a new df:

    .. code-block::

      df = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
      df.aggregate(np.hstack)

    Functions can be applied to grouped columns of multiple dataframes using
    `np.group_apply`.
    For example, the following applies stackes each column in each df horizontally:

    .. code-block::

      df1 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
      df2 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
      NumpyDataFrame.group_apply((df1, df2), np.hstack)

    In another example, we can apply np.divide to two dfs, using `expand=True` to expand
    the underlying arguments to properly run `np.divide`. The following two strategies
    are
    functionally equivalent:

    .. code-block::

      # strategy 1
      df1.group_apply((df1, df2), np.divide, expand=True)

      # strategy 2
      def div(a):
        return np.divide(a[0], a[1])
      df1.group_apply((df1, df2), div)

    **operations**

    Mathematical operations can be performed the same as you would np.ndarrays:

    .. code-block::

        df1 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
        df2 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})

        df3 = df1 + df2  # add each element in each column
        df3 += 10 # add 10 to each element
        df3 *= 2. # multiply each element by 2.
        df3 * df3 # multiply each element in each column element wise
        df3 ** 2
        df3 ** df2

    **concatenations and appending**

    Dataframes can be concatenated together by the following:

    .. code-block::

        df1 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
        df2 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
        NumpyDataFrame.concat((df1, df2))

    Dataframes with different column can be concatenated together by setting a fill
    value

    .. code-block::

        df1 = NumpyDataFrame({'A': np.arange(10), 'B': np.arange(10)})
        df2 = NumpyDataFrame({'A': np.arange(10), 'C': np.arange(10)})
        NumpyDataFrame.concat((df1, df2), fill_value=np.nan)

    **conversions**

    *to pandas*
    .. code-block::

        df1.to_df()

    *to numpy*
    .. code-block::

      df1.aggragate(np.hstack)

    *change dtype*

    .. code-block::
      df1.apply(np.astype, np.float64)
    """

    def __init__(self, data=None, apply=None):
        """Initializes a numpy data frame from a dict of string to np.ndarrays.
        The dict keys are representative of *column names* and the values are
        rows of that column. The shapes of each np.ndarray must be the same,
        else :class:`NumpyDataFrameException` is raised.

        :param data: A dict of string to np.ndarrays
        :param apply: Function to apply across the numpy data frame
        """
        if data is None:
            data = {}
        self._data = OrderedDict()
        self._data.update({k: v for k, v in data.items() if v is not None})
        if apply:
            self.apply(apply, inplace=True)
        self.validate()

    @property
    def data(self) -> Dict[str, np.ndarray]:
        """The underlying data dict of the dataframe."""
        return self._data

    @data.setter
    def data(self, new):
        """Set and validate the underlying data dict for the dataframe."""
        self._data = new
        self.validate()

    def validate(self):
        """Validate that the shapes of all of the np.ndarrays are the same."""
        shapes = {v.shape for v in self.data.values()}
        if len(shapes) > 1:
            keys_and_shapes = {}
            for k, v in self.data.items():
                keys_and_shapes.setdefault(v.shape, list()).append(k)
            raise NumpyDataFrameException(
                "{} can only have one shape. Found the following shapes {}. If you want"
                " to sqeeze all of the data, set 'apply=np.squeeze'".format(
                    self.__class__, keys_and_shapes
                )
            )
        for k, v in self.data.items():
            if not issubclass(type(v), np.ndarray):
                raise NumpyDataFrameException(
                    "{cls} only supports {typ}, not {x}".format(
                        cls=self.__class__.__name__, typ=np.ndarray, x=type(v)
                    )
                )

    def prefix(self, s: str, cols=None, inplace=False) -> "NumpyDataFrame":
        """Adds a prefix to all of the column names and returns a new
        dataframe."""
        if cols is None:
            cols = self.columns
        return self.apply_to_col_names(lambda x: s + x, cols=cols, inplace=inplace)

    def suffix(self, s: str, cols=None, inplace=False) -> "NumpyDataFrame":
        """Adds a prefix to all of the column names and returns a new
        dataframe."""
        if cols is None:
            cols = self.columns
        return self.apply_to_col_names(lambda x: x + s, cols=cols, inplace=inplace)

    def apply_to_col_names(
        self, func, *args, cols=None, inplace=False, **kwargs
    ) -> "NumpyDataFrame":
        """Apply a function to the column names and returns a new dataframe.

        :param func: the function to apply
        :param args: the additional function arguments
        :param cols: the columns to apply the function to. If False, all columns are
        used.
        :param inplace: if True, will apply the function to the current df and return
        the current df.
        :param kwargs: the additional function keyword arguments
        :return:
        """
        if cols is None:
            cols = self.columns
        data = {func(k, *args, **kwargs): v for k, v in self.data.items() if k in cols}
        if inplace:
            self.data = data
            return self
        return self.__class__(data)

    def aggregate(self, func, *args, cols=None, **kwargs) -> Any:
        """Group all of the np.ndarrays across all columns as a list and apply
        a function.

        :param func: the function to apply (e.g. np.hstack)
        :param args: the additional arguments of the function
        :param cols: the cols to apply the function to. If False, all columns are used.
        :param kwargs: the keyword arguments to apply to the function
        :return: the result of the function
        """
        if cols is None:
            cols = self.columns
        collapsed = [self.data[c] for c in cols]
        return func(collapsed, *args, **kwargs)

    def apply(
        self,
        func: Callable,
        *args,
        astype=None,
        preprocess=None,
        inplace=False,
        **kwargs,
    ) -> Any:
        """Apply a function to each np.ndarray.

        :param func: The function to appluy
        :param args: the additional arguments of the function
        :param astype: the type of data frame to return
        :param preprocess: preprocess function to apply to each np.ndarray before
        applying 'func'
        :param inplace: If True, will apply the function to the current df and return
        the current df.
        :param kwargs: the keyword arguments to apply to the function
        :return: a new dataframe
        """
        data = {}
        for k, v in self.data.items():
            try:
                if preprocess:
                    data[k] = func(preprocess(v), *args, **kwargs)
                else:
                    data[k] = func(v, *args, **kwargs)
            except Exception as e:
                raise NumpyDataFrameException(
                    "Could not apply '{}' because '{} {}'".format(
                        func.__name__, type(e), e
                    )
                ) from e
        if inplace:
            if astype is not None and astype is not self.__class__:
                raise NumpyDataFrameException(
                    "Cannot convert from {} to {} while inplace=True".format(
                        self.__class__, astype
                    )
                )
            self.data = data
            return self
        if astype is None:
            astype = self.__class__
        return astype(data)

    @classmethod
    def merge(cls, others: Iterable["NumpyDataFrame"]) -> "NumpyDataFrame":
        """Merge many dfs into a single df."""
        df = cls()
        for a in others:
            df.update(a)
        return df

    @classmethod
    def concat(
        cls, others: Iterable["NumpyDataFrame"], fill_value=Null
    ) -> "NumpyDataFrame":
        """Concatenate several dfs into a single df."""
        return cls.group_apply(others, np.hstack, _fill_value=fill_value)

    def append(self, other: "NumpyDataFrame") -> "NumpyDataFrame":
        """Append the contents of the other df to this df."""
        self.group_apply((other,), np.hstack)
        return self

    def fill_value(self, cols: Iterable[str], value: Any) -> None:
        """Create new columns, if they are missing, and fill them with the
        specified value."""
        for c in cols:
            if c not in self.col:
                self.col[c] = np.array([value for _ in range(len(self))])

    @classmethod
    def group_apply(
        cls,
        others: Iterable["NumpyDataFrame"],
        func,
        *args,
        expand=False,
        _fill_value=Null,
        **kwargs,
    ) -> "NumpyDataFrame":
        """Groups np.arrays according to their column name for several
        dataframes (as a list) and applies a function to each group. Returns a
        new df with the results.

        :param others: iterable of dfs
        :param func: the function to apply
        :param args: additional arguments for the function
        :param expand: If true, the list of np.arrays will be expanded, as in
        `func(*list_of_arrs, ...)`
        :param _fill_value:
        :param kwargs: additional keyword arguments for the function
        :return: a new df
        """
        d = {}
        other_cols = {tuple(sorted(o.columns)) for o in others}
        if len(other_cols) > 1:
            if _fill_value is Null:
                raise NumpyDataFrameException(
                    "Cannot apply to group. Different columns found: {}".format(
                        other_cols
                    )
                )
            else:
                all_cols = []
                for o in others:
                    all_cols += list(o.columns)
                all_cols = sorted(list(set(all_cols)))
                for o in others:
                    o.fill_value(all_cols, _fill_value)
        for o in others:
            for k, v in o.data.items():
                d.setdefault(k, list()).append(v)
        if expand:
            data = {k: func(*v, *args, **kwargs) for k, v in d.items()}
        else:
            data = {k: func(v, *args, **kwargs) for k, v in d.items()}
        return cls(data)

    @classmethod
    def stack(cls, others: Iterable["NumpyDataFrame"], axis):
        """Apply np.stack to each column for several dfs."""
        return cls.group_apply(others, np.stack, axis=axis)

    @classmethod
    def hstack(cls, others: Iterable["NumpyDataFrame"]):
        """Apply np.hstack to each column for several dfs."""
        return cls.group_apply(others, np.hstack)

    @classmethod
    def vstack(cls, others: List["NumpyDataFrame"]):
        """Apply np.vstack to each column for several dfs."""
        return cls.group_apply(others, np.vstack)

    @property
    def shape(self) -> Tuple[int, ...]:
        """Return the expected shape for the underlying np.ndarray.

        This is the shape of the array for any given column.
        """
        return list(self.data.values())[0].shape

    def reshape(self, shape) -> "NumpyDataFrame":
        """Reshape all arrays in the df."""
        return self.apply(np.reshape, shape)

    @property
    def columns(self) -> Tuple[str, ...]:
        """Return the column names."""
        return tuple(self.data)

    @property
    def col(self) -> "NumpyDataFrameIndexer":
        """Return the column indexer."""
        return NumpyDataFrameIndexer(self)

    def to_df(self, force=True) -> pd.DataFrame:
        """Force the NumpyDataFrame into a pandas.DataFrame."""
        print(self.shape)
        x = self
        if force:
            if len(x.shape) >= 2:
                x = x.apply(np.squeeze)
            elif len(x.shape) == 0:
                x = x.apply(np.expand_dims, axis=0)
        if len(x.shape) > 1:
            raise NumpyDataFrameException(
                "Unable to force dataframe of "
                "shape {} into a pandas.DataFrame".format(self.shape)
            )
        return pd.DataFrame(x.data)

    def update(self, data: Union["NumpyDataFrame", Dict[str, np.ndarray]], apply=None):
        """Update the df from a dict or another df."""
        if issubclass(type(data), NumpyDataFrame):
            data = data.data
        if issubclass(type(data), dict):
            self.data.update(data)
            if apply:
                self.apply(apply)
            self.validate()

    def items(self) -> ItemsView[str, np.ndarray]:
        """Iterate key: arr for the the underlying data dict."""
        return self.data.items()

    def copy(self) -> "NumpyDataFrame":
        """Copy the df."""
        return self.apply(np.copy)

    def dumps(self):
        """Use msgpack to dump df to a byte string."""
        return msgpack.dumps(self.data)

    def dump(self, f: str):
        """Dump byte repr of df to the specified path."""
        with open(f, "wb") as f:
            msgpack.dump(self.data, f)

    @classmethod
    def loads(cls, s: str) -> "NumpyDataFrame":
        """Use msgpack to load a df from a byte string."""
        data = msgpack.loads(s)
        data = {k.decode(): v for k, v in data.items()}
        return cls(data)

    @classmethod
    def load(cls, f: str) -> "NumpyDataFrame":
        """Load the byte repr of df from the specified path."""
        with open(f, "rb") as f:
            data = msgpack.load(f)
            data = {k.decode(): v for k, v in data.items()}
            return cls(data)

    def __getitem__(self, key: Union[int, slice, np.ndarray]) -> "NumpyDataFrame":
        new = self.__class__(self.data)
        new.data = {k: np.array(v[key]) for k, v in new.data.items()}
        return new

    def __setitem__(self, key: int, val: Any):
        for v in self.data.values():
            v[key] = val

    def __len__(self) -> int:
        return self.shape[0]

    def __iter__(self) -> Generator["NumpyDataFrame", None, None]:
        for i in range(len(self)):
            yield self[i]

    def __add__(self, other: "NumpyDataFrame") -> "NumpyDataFrame":
        if not issubclass(type(other), NumpyDataFrame):
            return self.apply(np.sum, preprocess=lambda x: (x, other))
        return self.stack([self, other], axis=1).apply(np.sum, axis=1)

    def __mul__(self, other: Union[int, float]) -> "NumpyDataFrame":
        if not issubclass(type(other), NumpyDataFrame):
            other = np.array([other] * self.shape[0])
            return self.apply(np.multiply, other)
        return self.group_apply([self, other], np.multiply, expand=True)

    def __truediv__(self, other: "NumpyDataFrame") -> "NumpyDataFrame":
        if not issubclass(type(other), NumpyDataFrame):
            other = np.array([other] * self.shape[0])
            return self.apply(np.divide, other)
        return self.group_apply([self, other], np.divide, expand=True)

    def __pow__(self, other: "NumpyDataFrame") -> "NumpyDataFrame":
        if not issubclass(type(other), NumpyDataFrame):
            other = np.array([other] * self.shape[0])
            return self.apply(np.power, other)
        return self.group_apply([self, other], np.power, expand=True)

    def __neg__(self) -> "NumpyDataFrame":
        return self * -1

    def __sub__(self, other: "NumpyDataFrame") -> "NumpyDataFrame":
        return self + -other

    def __str__(self) -> str:
        if self.shape == tuple():
            x = self.apply(np.expand_dims, axis=0)
            x = x.aggregate(np.stack, axis=0).flatten()
        else:
            x = self.aggregate(np.stack, axis=1)
        return "{}\ncols={}\n{}".format(self.__class__, self.columns, pprint.pformat(x))

    def __repr__(self) -> str:
        return "<{cls} shape={shape} cols={cols} dtype={dtype}>".format(
            cls=self.__class__.__name__,
            shape=self.shape,
            cols=self.columns,
            dtype=list(self.data.values())[0].dtype,
        )


class NumpyDataFrameIndexer(MappingABC):
    """The indexer for NumpyDataFrames."""

    def __init__(self, df: "NumpyDataFrame"):
        self.df = df

    def __len__(self) -> int:
        """Return number of columns."""
        return len(self.df.columns)

    def __iter__(self) -> Generator[str, None, None]:
        for c in self.df.columns:
            yield c

    def __contains__(self, item: str) -> bool:
        for c in self:
            if item == c:
                return True
        return False

    def __delitem__(self, key):
        del self.df.data[key]

    def __setitem__(self, col: str, val: np.ndarray):
        if not issubclass(type(val), np.ndarray):
            raise TypeError("Value must be a np.ndarray, not a {}".format(type(val)))
        self.df.data[col] = val
        self.df.validate()

    def __getitem__(self, cols: Union[str, Iterable[str]]) -> "NumpyDataFrame":
        if isinstance(cols, str):
            cols = (cols,)
        elif not isinstance(cols, IterableABC):
            cols = (cols,)
        data = {k: self.df.data[k] for k in self.df.data if k in cols}
        return self.df.__class__(data)



"""Module for calculating the complexity of a DNASequence.

Complexity rules borrowed from IDT.

* less than 25% GC content
* higher than 75% GC content
* homopolymeric runs of 10 or more A/Ts
* homopolymeric runs of 6 or more G/Cs
* repeats greater than 14bp
* hairpins
* GC scew?
* windowed scew?
"""
import re
from typing import Dict
from typing import Union

import numpy as np


# TODO: remove unnecessary code below
class ComplexityConfig:
    HOMOPOLYMERIC_AT = 10
    HOMOPOLYMERIC_GC = 6
    REPEAT_MAX_LENGTH = 14
    PARTITION_OVERLAP = 25


C = ComplexityConfig


at_pattern = re.compile("[AT]{" + str(C.HOMOPOLYMERIC_AT) + ",}")
gc_pattern = re.compile("[GC]{" + str(C.HOMOPOLYMERIC_GC) + ",}")


def get_AT_complexity(seq):
    cmax = 0
    for match in at_pattern.findall(seq, re.IGNORECASE):
        if len(match) >= C.HOMOPOLYMERIC_AT:
            c = len(match) - C.HOMOPOLYMERIC_AT + 1
            if c > cmax:
                cmax = c
    return cmax / 10.0


def get_GC_complexity(seq):
    cmax = 0
    for match in gc_pattern.findall(seq, re.IGNORECASE):
        if len(match) >= C.HOMOPOLYMERIC_GC:
            c = len(match) - C.HOMOPOLYMERIC_GC + 1
            if c > cmax:
                cmax = c
    return cmax / 10.0


def get_GC_content(seq):
    s = seq.upper()
    return (s.count("G") + s.count("C")) / len(s)


def get_GC_content_complexity(seq):
    gc = get_GC_content(seq)
    return abs(gc * 100.0 - 50) * 17 / 25.0


def iter_kmers(seq, length):
    for i in range(0, len(seq) - length):
        yield (i, seq[i : i + length])


def count_kmers(seq, length):
    kmers = {}
    for kmer in iter_kmers(seq, length):
        kmers.setdefault(kmer[1], list())
        kmers[kmer[1]].append(kmer[0])
    return kmers


def repeat_complexity_func(len_repeat, max_repeat_length):
    return (len_repeat - max_repeat_length) / 2.0 + 10


def get_repeat_complexity(seq, length):
    repeats = {}
    for kmer, positions in count_kmers(seq, length).items():
        if len(positions) > 1:
            kmer1 = kmer2 = kmer
            kmer_len = len(kmer)
            while kmer1 == kmer2:
                kmer_len += 1
                kmer1 = seq[positions[0] : positions[0] + kmer_len]
                kmer2 = seq[positions[1] : positions[1] + kmer_len]
            passes = True
            if len(kmer1) >= length:
                for k in repeats:
                    if kmer1 in k:
                        passes = False
                        break
                if passes:
                    repeats[kmer1] = repeat_complexity_func(len(kmer1), length)
    return repeats


def complexity(seq: str) -> Dict[str, Union[float, int]]:
    complexity_info = {}
    complexity_info["Homopolymeric AT"] = get_AT_complexity(seq)
    complexity_info["Homopolymeric GC"] = get_GC_complexity(seq)
    complexity_info["High GC Content"] = get_GC_content_complexity(seq)
    complexity_info["Repeats"] = get_repeat_complexity(seq, C.REPEAT_MAX_LENGTH)
    return complexity_info


def complexity_score(seq: str) -> float:
    data = complexity(seq)
    total = 0
    for k, v in data.items():
        if isinstance(v, int) or isinstance(v, float):
            total += v
        else:
            for _v in v.values():
                total += _v
    return total


# TODO: compute cost for extremes of GC content
# TODO: this could be a separate library
class DNAStats:
    """
    Class for computing statistics for DNA Sequences.

    ::

        {
            "n_repeats",
            "n_hairpins",
            "window_cost",
            "gc_cost"
        }
    """

    BASES = "AGCT"
    np.random.seed(1)
    BASE_SIGNATURES = np.random.uniform(0.0, 1.0, size=len(BASES)).reshape(-1, 4)

    def __init__(self, seq, repeat_window, stats_window, hairpin_window):
        self.seq = seq
        arr = np.array(list(seq))
        self.bases = self.BASES

        self.seq_onehot = self.one_hot(arr, self.bases)
        self.repeat_window = repeat_window
        self.stats_window = stats_window
        self.repeat_signatures = self.get_repeat_signatures(repeat_window)

        rolling_stats = self.get_base_stats(stats_window)
        gc_content = (
            rolling_stats[self.BASES.index("G"), :]
            + rolling_stats[self.BASES.index("C"), :]
        )
        at_content = (
            rolling_stats[self.BASES.index("A"), :]
            + rolling_stats[self.BASES.index("T"), :]
        )
        self.rolling_stats = np.vstack((rolling_stats, gc_content, at_content))

        self.fwd_signatures, self.rev_signatures = self.get_hairpin_signatures(
            hairpin_window
        )
        self.cached_partitions = []

    @staticmethod
    def one_hot(sequence, categories):
        arrs = []
        for i, c in enumerate(categories):
            a = sequence == c
            a = a.astype(int)
            arrs.append(a)
        return np.vstack(arrs)

    @staticmethod
    def rolling_window(a, window):
        shape = a.shape[:-1] + (a.shape[-1] - window + 1, window)
        strides = a.strides + (a.strides[-1],)
        return np.lib.stride_tricks.as_strided(a, shape=shape, strides=strides)

    @staticmethod
    def rolling_average(x, n):
        mv = np.convolve(x, np.ones(n) / n, mode="valid")
        return np.concatenate(([np.NaN for k in range(n - 1)], mv))

    def get_base_stats(self, window):
        return np.mean(self.rolling_window(self.seq_onehot, window), axis=2)

    def get_repeat_signatures(self, window):
        seq_signatures = np.sum(
            np.multiply(self.seq_onehot, self.BASE_SIGNATURES.T), axis=0
        )
        x = np.random.uniform(0.0, 100.0, size=window)
        mv = np.convolve(seq_signatures, x, mode="valid")
        signatures = np.concatenate(([np.NaN for _ in range(window - 1)], mv))
        return signatures

    def get_hairpin_signatures(self, window):
        seq_signatures = np.sum(
            np.multiply(self.seq_onehot, self.BASE_SIGNATURES.T), axis=0
        )
        revcomp_signatures = np.sum(
            np.multiply(self.seq_onehot[:, ::-1], self.BASE_SIGNATURES[:, ::-1].T),
            axis=0,
        )
        x = np.random.uniform(0.0, 100.0, size=window)
        mv1 = np.convolve(seq_signatures, x, mode="valid")
        mv2 = np.convolve(revcomp_signatures, x, mode="valid")
        return mv1, mv2[::-1]

    @staticmethod
    def count_signatures(a):
        return len(np.where(np.unique(a, return_counts=True)[1] > 1)[0])

    def slice_repeats(self, i, j):
        return self.count_signatures(self.repeat_signatures[i:j])

    def slice_hairpins(self, i, j):
        f = self.fwd_signatures[i:j]
        r = self.rev_signatures[i:j]
        d = np.concatenate((f, r))
        num_hairpins = self.count_signatures(d) - 2 * self.count_signatures(f)
        return num_hairpins

    def window_cost(self, i, j):
        d = self.rolling_stats[:4, i:j]

        # keep each base below 0.8
        a = np.sum(d > 0.75, axis=1)

        # keep gc content below extremes 0.7
        b = np.sum(self.rolling_stats[4, i:j] > 0.85)
        return a, b

    # def get_GC_content_complexity(seq):
    #     gc = get_GC_content(seq)
    #     return abs(gc * 100.0 - 50) * 17 / 25.0

    @staticmethod
    def _optimize_partition_helper(
        signatures: np.ndarray, step: int, i: int = None, j: int = None
    ):
        """Optimize partition by minimizing the number of signatures in the
        given array.

        :param signatures: array of signatures
        :param step: step size
        :param i:
        :param j:
        :return:
        """
        d = []

        if i is None:
            i = 0
        if j is None:
            j = signatures.shape[1]

        for x in range(i, j, step):
            m1 = np.empty(signatures.shape[1])
            m2 = m1.copy()
            m1.fill(np.nan)
            m2.fill(np.nan)

            m1[:x] = np.random.uniform(1, 10)
            m2[x:] = np.random.uniform(1, 10)

            d += [m1, m2]
        d = np.vstack(d)
        z = np.tile(d, signatures.shape[0]) * signatures.flatten()

        partition_index = np.repeat(
            np.arange(0, signatures.shape[1], step),
            signatures.shape[0] * signatures.shape[1] * 2,
        )

        a, b, c = np.unique(z, return_counts=True, return_index=True)
        i = b[np.where(c > 1)]
        a, c = np.unique(partition_index[i], return_counts=True)
        if len(c):
            arg = c.argmin()
            return a[arg], c[arg]

    def partition(
        self,
        step,
        overlap,
        i=None,
        j=None,
        border=100,
        stopping_threshold=10,
        window=None,
        use_cache=True,
    ):
        if j is None:
            j = len(self.seq)
        if i is None:
            i = 0

        if use_cache:
            for cached in self.cached_partitions:
                if i >= cached[0] and j <= cached[1]:
                    if cached[2]["cost"] < stopping_threshold:
                        window = (
                            cached[2]["index_1"][1] - overlap,
                            cached[2]["index_2"][0] + overlap,
                        )
                elif i <= cached[0] and j >= cached[1]:
                    if cached[2]["cost"] > stopping_threshold:
                        return []

        if not window:
            window = (i, j)

        to_search = []
        for index in range(window[0], window[1], step):
            i1 = index
            i2 = index - overlap
            if i2 <= border:
                continue
            if i1 > len(self.seq) - border:
                continue
            to_search.append((i1, i2))

        if not to_search:
            return []

        search_index = int(len(to_search) / 2)
        searched_costs = []
        searched = []
        while search_index not in searched:

            searched.append(search_index)

            try:
                i1, i2 = to_search[search_index]
            except IndexError:
                print(search_index)
                print(len(to_search))
                raise IndexError

            c1 = self.cost(None, i1)
            c2 = self.cost(i2, None)
            c = c1 + c2

            searched_costs.append((c, c1, c2, i1, i2, search_index))

            if c1 > c2:
                search_index = int(search_index / 2)
            elif c2 > c1:
                search_index = int((len(to_search) - search_index) / 2) + search_index

            if c < stopping_threshold:
                break

        searched_costs = sorted(searched_costs)
        c, c1, c2, i1, i2, search_index = searched_costs[0]

        partitions = self.partition_scan(
            1,
            overlap=overlap,
            i=i,
            j=j,
            window=(min(i1, i2) - overlap, max(i1, i2) + overlap),
        )
        if use_cache:
            self.cached_partitions.append((i, j, partitions[0]))
        return partitions

    def partition_scan(self, step, overlap, i=None, j=None, window=None):
        if j is None:
            j = len(self.seq)
        if i is None:
            i = 0

        if not window:
            window = (i, j)

        costs = []
        for x in range(window[0], window[1], step):
            i2, j2 = x, x - overlap

            c1 = self.cost(i, i2)
            c2 = self.cost(j2, j)
            costs.append(
                {
                    "cost": c1 + c2,
                    "cost_1": c1,
                    "cost_2": c2,
                    "index_1": (i, i2),
                    "index_2": (j2, j),
                }
            )
        costs = sorted(costs, key=lambda x: x["cost"])
        return costs

    def cost(self, i=None, j=None):
        d = self(i, j)
        w1 = d["window_cost"][0]
        w2 = d["window_cost"][1]
        total = (
            d["n_repeats"] + d["n_hairpins"] + np.sum(w1) + np.sum(w2) + d["gc_cost"]
        )
        return total

    def __call__(self, i=None, j=None):
        mn = np.mean(self.seq_onehot[:, i:j], axis=1)
        gi = self.bases.index("G")
        ci = self.bases.index("C")
        gc = mn[gi] + mn[ci]
        gccost = abs(gc * 100.0 - 50) * 17 / 25.0
        return {
            "n_repeats": self.slice_repeats(i, j),
            "n_hairpins": self.slice_hairpins(i, j),
            "window_cost": self.window_cost(i, j),
            "gc_cost": gccost,
        }
